# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

# risks from classes.dex :
    # API <-> Permissions 
        # method X is more dangerous than another one
    # const-string -> apk-tool
        # v0 <- X
        # v1 <- Y

        # v10 <- X
        # v11 <- Y

        # CALL( v0, v1 )
    # obfuscated names

GENERAL_RISK            = 0
DANGEROUS_RISK          = 1
SIGNATURE_SYSTEM_RISK   = 2 
SIGNATURE_RISK          = 3
NORMAL_RISK             = 4

MONEY_RISK              = 5 
SMS_RISK                = 6
PHONE_RISK              = 7
INTERNET_RISK           = 8
PRIVACY_RISK            = 9 
DYNAMIC_RISK            = 10

BINARY_RISK             = 11
EXPLOIT_RISK            = 12

RISK_VALUES = {
    DANGEROUS_RISK          : 5,
    SIGNATURE_SYSTEM_RISK   : 10,
    SIGNATURE_RISK          : 10,
    NORMAL_RISK             : 0,

    MONEY_RISK              : 5,
    SMS_RISK                : 5,
    PHONE_RISK              : 5,
    INTERNET_RISK           : 2,
    PRIVACY_RISK            : 5,
    DYNAMIC_RISK            : 5,

    BINARY_RISK             : 10,
    EXPLOIT_RISK            : 15,
}

GENERAL_PERMISSIONS_RISK = {
    "dangerous"                 : DANGEROUS_RISK,
    "signatureOrSystem"         : SIGNATURE_SYSTEM_RISK,
    "signature"                 : SIGNATURE_RISK,
    "normal"                    : NORMAL_RISK,
}

PERMISSIONS_RISK = {
    "SEND_SMS"                  : [ MONEY_RISK, SMS_RISK ],

    "RECEIVE_SMS"               : [ SMS_RISK ],
    "READ_SMS"                  : [ SMS_RISK ],
    "WRITE_SMS"                 : [ SMS_RISK ],
    "RECEIVE_SMS"               : [ SMS_RISK ],
    "RECEIVE_MMS"               : [ SMS_RISK ],

    "PHONE_CALL"                : [ MONEY_RISK ],
    "PROCESS_OUTGOING_CALLS"    : [ MONEY_RISK ],
    "CALL_PRIVILEGED"           : [ MONEY_RISK ],


    "INTERNET"                  : [ MONEY_RISK, INTERNET_RISK ],

    "READ_PHONE_STATE"          : [ PRIVACY_RISK ],
    "READ_CONTACTS"             : [ PRIVACY_RISK ],
    "READ_HISTORY_BOOKMARKS"    : [ PRIVACY_RISK ],
    "ACCESS_FINE_LOCATION"      : [ PRIVACY_RISK ],
    "ACCESS_COARSE_LOCATION"    : [ PRIVACY_RISK ],
}

LOW_RISK                    = "low"
AVERAGE_RISK                = "average"
HIGH_RISK                   = "high"
UNACCEPTABLE_RISK           = "unacceptable"

NULL_MALWARE_RISK           = "null"
AVERAGE_MALWARE_RISK        = "average"
HIGH_MALWARE_RISK           = "high"
UNACCEPTABLE_MALWARE_RISK   = "unacceptable"

from androguard.core.androconf import error, warning, debug, set_debug, get_debug
from androguard.core.bytecodes import dvm
from androguard.core.analysis import analysis

from androguard.core.bytecodes.dvm_permissions import DVM_PERMISSIONS
import re, copy

def add_system_rule(system, rule_name, rule) :
    system.rules[ rule_name ] = rule

def create_system_risk() :
    try :
        import fuzzy
    except ImportError :
        error("please install pyfuzzy to use this module !")

    import fuzzy.System
    import fuzzy.InputVariable
    import fuzzy.fuzzify.Plain
    import fuzzy.OutputVariable
    import fuzzy.defuzzify.COGS
    import fuzzy.defuzzify.COG
    import fuzzy.defuzzify.MaxRight
    import fuzzy.defuzzify.MaxLeft
    import fuzzy.defuzzify.LM
    import fuzzy.set.Polygon
    import fuzzy.set.Singleton
    import fuzzy.set.Triangle
    import fuzzy.Adjective
    import fuzzy.operator.Input
    import fuzzy.operator.Compound
    import fuzzy.norm.Min
    import fuzzy.norm.Max
    import fuzzy.Rule
    import fuzzy.defuzzify.Dict

    system = fuzzy.System.System()

    input_Dangerous_Risk = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_Money_Risk = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_Privacy_Risk = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_Binary_Risk = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_Internet_Risk = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_Dynamic_Risk = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    
    # Input variables

        # Dangerous Risk
    system.variables["input_Dangerous_Risk"] = input_Dangerous_Risk
    input_Dangerous_Risk.adjectives[LOW_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (8.0, 1.0), (12.0, 0.0)]) )
    input_Dangerous_Risk.adjectives[AVERAGE_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(8.0, 0.0), (50.0, 1.0), (60.0, 0.0)]) )
    input_Dangerous_Risk.adjectives[HIGH_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(50.0, 0.0), (85.0, 1.0), (95.0, 0.0)]) )
    input_Dangerous_Risk.adjectives[UNACCEPTABLE_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(85.0, 0.0), (100.0, 1.0)]) )

        # Money Risk
    system.variables["input_Money_Risk"] = input_Money_Risk
    input_Money_Risk.adjectives[LOW_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (2.0, 1.0), (3.0, 0.0)]) )
    input_Money_Risk.adjectives[UNACCEPTABLE_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(4.0, 0.0), (5.0, 1.0), (30.0, 1.0)]) )

        # Privacy Risk
    system.variables["input_Privacy_Risk"] = input_Privacy_Risk
    input_Privacy_Risk.adjectives[LOW_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (6.0, 1.0), (10.0, 0.0)]) )
    input_Privacy_Risk.adjectives[HIGH_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(6.0, 0.0), (10.0, 1.0), (20.0, 0.0)]) )
    input_Privacy_Risk.adjectives[UNACCEPTABLE_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(15.0, 0.0), (20.0, 1.0), (30.0, 1.0)]) )
    
        # Binary Risk
    system.variables["input_Binary_Risk"] = input_Binary_Risk
    input_Binary_Risk.adjectives[LOW_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (6.0, 1.0), (10.0, 0.0)]) )
    input_Binary_Risk.adjectives[AVERAGE_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(6.0, 0.0), (10.0, 1.0), (15.0, 0.0)]) )
    input_Binary_Risk.adjectives[HIGH_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(10.0, 0.0), (20.0, 1.0), (24.0, 0.0)]) )
    input_Binary_Risk.adjectives[UNACCEPTABLE_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(23.0, 0.0), (30.0, 1.0), (40.0, 1.0)]) )

        # Internet Risk
    system.variables["input_Internet_Risk"] = input_Internet_Risk
    #input_Internet_Risk.adjectives[LOW_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (1.0, 1.0)]) )
    input_Internet_Risk.adjectives[HIGH_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(1.0, 0.0), (5.0, 1.0), (30.0, 1.0)]) )

        # Dynamic Risk
    system.variables["input_Dynamic_Risk"] = input_Dynamic_Risk
    input_Dynamic_Risk.adjectives[LOW_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (2.0, 1.0), (3.0, 0.0)]))
    input_Dynamic_Risk.adjectives[UNACCEPTABLE_RISK] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(4.0, 0.0), (5.0, 1.0), (50.0, 1.0)]) )

    # Output variables
    output_malware_risk = fuzzy.OutputVariable.OutputVariable(
                            defuzzify=fuzzy.defuzzify.COGS.COGS(),
                            description="malware risk",
                            min=0.0,max=100.0,
                        )
    
    #output_malware_risk = fuzzy.OutputVariable.OutputVariable(defuzzify=fuzzy.defuzzify.Dict.Dict())

    output_malware_risk.adjectives[NULL_MALWARE_RISK] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(0.0))
    output_malware_risk.adjectives[AVERAGE_MALWARE_RISK] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(30.0))
    output_malware_risk.adjectives[HIGH_MALWARE_RISK] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(60.0))
    output_malware_risk.adjectives[UNACCEPTABLE_MALWARE_RISK] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(100.0))
    
    system.variables["output_malware_risk"] = output_malware_risk

    # Rules
    #RULE 0: DYNAMIC 
    add_system_rule(system, "r0", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[NULL_MALWARE_RISK]],
                                        operator=fuzzy.operator.Input.Input( system.variables["input_Dynamic_Risk"].adjectives[LOW_RISK] )
                    )
    )
    
    add_system_rule(system, "r0a", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[UNACCEPTABLE_MALWARE_RISK]],
                                        operator=fuzzy.operator.Input.Input( system.variables["input_Dynamic_Risk"].adjectives[UNACCEPTABLE_RISK] )
                    )
    )
    

    #RULE 1: MONEY
    add_system_rule(system, "r1", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[NULL_MALWARE_RISK]],
                                        operator=fuzzy.operator.Input.Input( system.variables["input_Money_Risk"].adjectives[LOW_RISK] )
                    )
    )

    add_system_rule(system, "r1a", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[UNACCEPTABLE_MALWARE_RISK]],
                                        operator=fuzzy.operator.Input.Input( system.variables["input_Money_Risk"].adjectives[UNACCEPTABLE_RISK] )
                    )
    )
    
    #RULE 3 : BINARY
    add_system_rule(system, "r3", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[AVERAGE_MALWARE_RISK]],
                                        operator=fuzzy.operator.Input.Input( system.variables["input_Binary_Risk"].adjectives[AVERAGE_RISK] )
                    )
    )
    
    add_system_rule(system, "r3a", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[HIGH_RISK]],
                                        operator=fuzzy.operator.Input.Input( system.variables["input_Binary_Risk"].adjectives[HIGH_RISK] )
                    )
    )

    add_system_rule(system, "r3b", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[UNACCEPTABLE_MALWARE_RISK]],
                                        operator=fuzzy.operator.Input.Input( system.variables["input_Binary_Risk"].adjectives[UNACCEPTABLE_RISK] )
                    )
    )
    
    # PRIVACY + INTERNET
    add_system_rule(system, "r5", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[HIGH_MALWARE_RISK]],
                                        operator=fuzzy.operator.Compound.Compound(
                                            fuzzy.norm.Min.Min(),
                                            fuzzy.operator.Input.Input( system.variables["input_Privacy_Risk"].adjectives[LOW_RISK] ),
                                            fuzzy.operator.Input.Input( system.variables["input_Internet_Risk"].adjectives[HIGH_RISK] ) )
                    )
    )
    add_system_rule(system, "r5a", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[UNACCEPTABLE_MALWARE_RISK]],
                                        operator=fuzzy.operator.Compound.Compound(
                                            fuzzy.norm.Min.Min(),
                                            fuzzy.operator.Input.Input( system.variables["input_Privacy_Risk"].adjectives[HIGH_RISK] ),
                                            fuzzy.operator.Input.Input( system.variables["input_Internet_Risk"].adjectives[HIGH_RISK] ) )
                    )
    )
    
    add_system_rule(system, "r6", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[HIGH_RISK]],
                                        operator=fuzzy.operator.Input.Input( system.variables["input_Dangerous_Risk"].adjectives[HIGH_RISK] )
                    )
    )
    
    add_system_rule(system, "r6a", fuzzy.Rule.Rule(
                                        adjective=[system.variables["output_malware_risk"].adjectives[UNACCEPTABLE_RISK]],
                                        operator=fuzzy.operator.Input.Input( system.variables["input_Dangerous_Risk"].adjectives[UNACCEPTABLE_RISK] )
                    )
    )


    return system


PERFECT_SCORE               = "perfect"
HIGH_SCORE                  = "high"
AVERAGE_SCORE               = "average"
LOW_SCORE                   = "low"
NULL_METHOD_SCORE           = "null"
AVERAGE_METHOD_SCORE        = "average"
HIGH_METHOD_SCORE           = "high"
PERFECT_METHOD_SCORE        = "perfect"

def create_system_method_score() :
    try :
        import fuzzy
    except ImportError :
        error("please install pyfuzzy to use this module !")

    import fuzzy.System
    import fuzzy.InputVariable
    import fuzzy.fuzzify.Plain
    import fuzzy.OutputVariable
    import fuzzy.defuzzify.COGS
    import fuzzy.set.Polygon
    import fuzzy.set.Singleton
    import fuzzy.set.Triangle
    import fuzzy.Adjective
    import fuzzy.operator.Input
    import fuzzy.operator.Compound
    import fuzzy.norm.Min
    import fuzzy.norm.Max
    import fuzzy.Rule
    
    system = fuzzy.System.System()

    input_Length_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_Match_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_AndroidEntropy_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_JavaEntropy_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_Permissions_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_Similarity_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    
    # Input variables

        # Length 
    system.variables["input_Length_MS"] = input_Length_MS
    input_Length_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (50.0, 1.0), (100.0, 0.0)]) )
    input_Length_MS.adjectives[AVERAGE_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(50.0, 0.0), (100.0, 1.0), (150.0, 1.0), (300.0, 0.0)]) )
    input_Length_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(150.0, 0.0), (200.0, 1.0), (300.0, 1.0), (400.0, 0.0)]) )
    input_Length_MS.adjectives[PERFECT_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(350.0, 0.0), (400.0, 1.0), (500.0, 1.0)]) )

        # Match
    system.variables["input_Match_MS"] = input_Match_MS
    input_Match_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (20.0, 1.0), (50.0, 0.0)]) )
    input_Match_MS.adjectives[AVERAGE_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(40.0, 0.0), (45.0, 1.0), (60.0, 1.0), (80.0, 0.0)]) )
    input_Match_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(75.0, 0.0), (90.0, 1.0), (98.0, 1.0), (99.0, 0.0)]) )
    input_Match_MS.adjectives[PERFECT_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(98.0, 0.0), (100.0, 1.0)]) )
    #input_Match_MS.adjectives[PERFECT_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Singleton.Singleton( 100.0 ) )

        # Android Entropy
    system.variables["input_AndroidEntropy_MS"] = input_AndroidEntropy_MS
    input_AndroidEntropy_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (2.0, 1.0), (4.0, 0.0)]) )
    input_AndroidEntropy_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(3.0, 0.0), (4.0, 1.0), (30.0, 1.0)]) )

        # Java Entropy
    system.variables["input_JavaEntropy_MS"] = input_JavaEntropy_MS
    input_JavaEntropy_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (2.0, 1.0), (4.0, 0.0)]) )
    input_JavaEntropy_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(3.0, 0.0), (4.0, 1.0), (30.0, 1.0)]) )
    
        # Permissions
    system.variables["input_Permissions_MS"] = input_Permissions_MS
    input_Permissions_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (3.0, 1.0), (4.0, 0.0)]) )
    input_Permissions_MS.adjectives[AVERAGE_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(3.0, 0.0), (4.0, 1.0), (8.0, 1.0), (9.0, 0.0)]) )
    input_Permissions_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(8.0, 0.0), (10.0, 1.0), (12.0, 1.0), (13.0, 0.0)]) )
    input_Permissions_MS.adjectives[PERFECT_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(12.0, 0.0), (13.0, 1.0), (20.0, 1.0)]) )
    
        # Similarity Match 
    system.variables["input_Similarity_MS"] = input_Similarity_MS
    input_Similarity_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (0.1, 1.0), (0.3, 0.0)]) )
    input_Similarity_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.3, 0.0), (0.35, 1.0), (0.4, 1.0)]) )
    

    # Output variables
    output_method_score = fuzzy.OutputVariable.OutputVariable(
                                defuzzify=fuzzy.defuzzify.COGS.COGS(),
                                description="method score",
                                min=0.0,max=100.0,
                             )
    output_method_score.adjectives[NULL_METHOD_SCORE] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(0.0))
    output_method_score.adjectives[AVERAGE_METHOD_SCORE] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(50.0))
    output_method_score.adjectives[HIGH_METHOD_SCORE] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(80.0))
    output_method_score.adjectives[PERFECT_METHOD_SCORE] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(100.0))

    system.variables["output_method_score"] = output_method_score
    
    add_system_rule(system, "android entropy null", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_score"].adjectives[NULL_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_AndroidEntropy_MS"].adjectives[LOW_SCORE] ))
    )
    
    add_system_rule(system, "java entropy null", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_score"].adjectives[NULL_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_JavaEntropy_MS"].adjectives[LOW_SCORE] ))
    )
    
    add_system_rule(system, "permissions null", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_score"].adjectives[NULL_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_Permissions_MS"].adjectives[LOW_SCORE] ))
    )
    
    add_system_rule(system, "permissions average", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_score"].adjectives[AVERAGE_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_Permissions_MS"].adjectives[AVERAGE_SCORE] ))
    )
    
    add_system_rule(system, "permissions high", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_score"].adjectives[HIGH_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_Permissions_MS"].adjectives[HIGH_SCORE] ))
    )
   
    add_system_rule(system, "permissions perfect", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_score"].adjectives[PERFECT_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_Permissions_MS"].adjectives[PERFECT_SCORE] ))
    )
   
    add_system_rule(system, "similarity low", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_score"].adjectives[NULL_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_Similarity_MS"].adjectives[LOW_SCORE] ))
    )

    add_system_rule(system, "length match perfect", fuzzy.Rule.Rule(
                                    adjective=[system.variables["output_method_score"].adjectives[PERFECT_METHOD_SCORE]],
                                    operator=fuzzy.operator.Compound.Compound(
                                        fuzzy.norm.Min.Min(),
                                        fuzzy.operator.Input.Input( system.variables["input_Length_MS"].adjectives[PERFECT_SCORE] ),
                                        fuzzy.operator.Input.Input( system.variables["input_Match_MS"].adjectives[PERFECT_SCORE] ) )
                                    )
    )

    add_system_rule(system, "length match null", fuzzy.Rule.Rule(
                                    adjective=[system.variables["output_method_score"].adjectives[NULL_METHOD_SCORE]],
                                    operator=fuzzy.operator.Compound.Compound(
                                        fuzzy.norm.Min.Min(),
                                        fuzzy.operator.Input.Input( system.variables["input_Length_MS"].adjectives[LOW_SCORE] ),
                                        fuzzy.operator.Input.Input( system.variables["input_Match_MS"].adjectives[PERFECT_SCORE] ) )
                                    )
    )

    add_system_rule(system, "length AndroidEntropy perfect", fuzzy.Rule.Rule(
                                    adjective=[system.variables["output_method_score"].adjectives[HIGH_METHOD_SCORE]],
                                    operator=fuzzy.operator.Compound.Compound(
                                        fuzzy.norm.Min.Min(),
                                        fuzzy.operator.Input.Input( system.variables["input_Length_MS"].adjectives[PERFECT_SCORE] ),
                                        fuzzy.operator.Input.Input( system.variables["input_AndroidEntropy_MS"].adjectives[HIGH_SCORE] ) )
                                    )
    )
    
    add_system_rule(system, "length JavaEntropy perfect", fuzzy.Rule.Rule(
                                    adjective=[system.variables["output_method_score"].adjectives[HIGH_METHOD_SCORE]],
                                    operator=fuzzy.operator.Compound.Compound(
                                        fuzzy.norm.Min.Min(),
                                        fuzzy.operator.Input.Input( system.variables["input_Length_MS"].adjectives[PERFECT_SCORE] ),
                                        fuzzy.operator.Input.Input( system.variables["input_JavaEntropy_MS"].adjectives[HIGH_SCORE] ) )
                                    )
    )


    add_system_rule(system, "length similarity perfect", fuzzy.Rule.Rule(
                                    adjective=[system.variables["output_method_score"].adjectives[PERFECT_METHOD_SCORE]],
                                    operator=fuzzy.operator.Compound.Compound(
                                        fuzzy.norm.Min.Min(),
                                        fuzzy.operator.Input.Input( system.variables["input_Length_MS"].adjectives[PERFECT_SCORE] ),
                                        fuzzy.operator.Input.Input( system.variables["input_Similarity_MS"].adjectives[HIGH_SCORE] ),
                                        )
                                    )
    )

    add_system_rule(system, "length similarity average", fuzzy.Rule.Rule(
                                    adjective=[system.variables["output_method_score"].adjectives[HIGH_METHOD_SCORE]],
                                    operator=fuzzy.operator.Compound.Compound(
                                        fuzzy.norm.Min.Min(),
                                        fuzzy.operator.Input.Input( system.variables["input_Length_MS"].adjectives[AVERAGE_SCORE] ),
                                        fuzzy.operator.Input.Input( system.variables["input_Similarity_MS"].adjectives[HIGH_SCORE] ),
                                        )
                                    )
    )

    return system

def create_system_method_one_score() :
    try :
        import fuzzy
    except ImportError :
        error("please install pyfuzzy to use this module !")

    import fuzzy.System
    import fuzzy.InputVariable
    import fuzzy.fuzzify.Plain
    import fuzzy.OutputVariable
    import fuzzy.defuzzify.COGS
    import fuzzy.set.Polygon
    import fuzzy.set.Singleton
    import fuzzy.set.Triangle
    import fuzzy.Adjective
    import fuzzy.operator.Input
    import fuzzy.operator.Compound
    import fuzzy.norm.Min
    import fuzzy.norm.Max
    import fuzzy.Rule
    
    system = fuzzy.System.System()

    input_Length_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_AndroidEntropy_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_JavaEntropy_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    input_Permissions_MS = fuzzy.InputVariable.InputVariable(fuzzify=fuzzy.fuzzify.Plain.Plain())
    
    # Input variables

        # Length 
    system.variables["input_Length_MS"] = input_Length_MS
    input_Length_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (50.0, 1.0), (100.0, 0.0)]) )
    input_Length_MS.adjectives[AVERAGE_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(50.0, 0.0), (100.0, 1.0), (150.0, 1.0), (300.0, 0.0)]) )
    input_Length_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(150.0, 0.0), (200.0, 1.0), (300.0, 1.0), (400.0, 0.0)]) )
    input_Length_MS.adjectives[PERFECT_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(350.0, 0.0), (400.0, 1.0), (500.0, 1.0)]) )

        # Android Entropy
    system.variables["input_AndroidEntropy_MS"] = input_AndroidEntropy_MS
    input_AndroidEntropy_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (2.0, 1.0), (4.0, 0.0)]) )
    input_AndroidEntropy_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(3.0, 0.0), (4.0, 1.0), (30.0, 1.0)]) )

        # Java Entropy
    system.variables["input_JavaEntropy_MS"] = input_JavaEntropy_MS
    input_JavaEntropy_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (2.0, 1.0), (4.0, 0.0)]) )
    input_JavaEntropy_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(3.0, 0.0), (4.0, 1.0), (30.0, 1.0)]) )
    
        # Permissions
    system.variables["input_Permissions_MS"] = input_Permissions_MS
    input_Permissions_MS.adjectives[LOW_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(0.0, 1.0), (3.0, 1.0), (4.0, 0.0)]) )
    input_Permissions_MS.adjectives[AVERAGE_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(3.0, 0.0), (4.0, 1.0), (8.0, 1.0), (9.0, 0.0)]) )
    input_Permissions_MS.adjectives[HIGH_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(8.0, 0.0), (10.0, 1.0), (12.0, 1.0), (13.0, 0.0)]) )
    input_Permissions_MS.adjectives[PERFECT_SCORE] = fuzzy.Adjective.Adjective( fuzzy.set.Polygon.Polygon([(12.0, 0.0), (13.0, 1.0), (20.0, 1.0)]) )
    
    # Output variables
    output_method_score = fuzzy.OutputVariable.OutputVariable(
                                defuzzify=fuzzy.defuzzify.COGS.COGS(),
                                description="method one score",
                                min=0.0,max=100.0,
                             )
    output_method_score.adjectives[NULL_METHOD_SCORE] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(0.0))
    output_method_score.adjectives[AVERAGE_METHOD_SCORE] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(50.0))
    output_method_score.adjectives[HIGH_METHOD_SCORE] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(80.0))
    output_method_score.adjectives[PERFECT_METHOD_SCORE] = fuzzy.Adjective.Adjective(fuzzy.set.Singleton.Singleton(100.0))

    system.variables["output_method_one_score"] = output_method_score
    
    add_system_rule(system, "android entropy null", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_one_score"].adjectives[NULL_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_AndroidEntropy_MS"].adjectives[LOW_SCORE] ))
    )
    
    add_system_rule(system, "java entropy null", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_one_score"].adjectives[NULL_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_JavaEntropy_MS"].adjectives[LOW_SCORE] ))
    )
    
    add_system_rule(system, "permissions null", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_one_score"].adjectives[NULL_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_Permissions_MS"].adjectives[LOW_SCORE] ))
    )
    
    add_system_rule(system, "permissions average", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_one_score"].adjectives[AVERAGE_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_Permissions_MS"].adjectives[AVERAGE_SCORE] ))
    )
    
    add_system_rule(system, "permissions high", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_one_score"].adjectives[HIGH_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_Permissions_MS"].adjectives[HIGH_SCORE] ))
    )
   
    add_system_rule(system, "permissions perfect", fuzzy.Rule.Rule(
                                                        adjective=[system.variables["output_method_one_score"].adjectives[PERFECT_METHOD_SCORE]],
                                                        operator=fuzzy.operator.Input.Input( system.variables["input_Permissions_MS"].adjectives[PERFECT_SCORE] ))
    )
   

    add_system_rule(system, "length permissions perfect", fuzzy.Rule.Rule(
                                    adjective=[system.variables["output_method_one_score"].adjectives[PERFECT_METHOD_SCORE]],
                                    operator=fuzzy.operator.Compound.Compound(
                                        fuzzy.norm.Min.Min(),
                                        fuzzy.operator.Input.Input( system.variables["input_Length_MS"].adjectives[PERFECT_SCORE] ),
                                        fuzzy.operator.Input.Input( system.variables["input_Permissions_MS"].adjectives[PERFECT_SCORE] ) )
                                    )
    )

    add_system_rule(system, "length AndroidEntropy perfect", fuzzy.Rule.Rule(
                                    adjective=[system.variables["output_method_one_score"].adjectives[HIGH_METHOD_SCORE]],
                                    operator=fuzzy.operator.Compound.Compound(
                                        fuzzy.norm.Min.Min(),
                                        fuzzy.operator.Input.Input( system.variables["input_Length_MS"].adjectives[PERFECT_SCORE] ),
                                        fuzzy.operator.Input.Input( system.variables["input_AndroidEntropy_MS"].adjectives[HIGH_SCORE] ) )
                                    )
    )
    
    add_system_rule(system, "length JavaEntropy perfect", fuzzy.Rule.Rule(
                                    adjective=[system.variables["output_method_one_score"].adjectives[HIGH_METHOD_SCORE]],
                                    operator=fuzzy.operator.Compound.Compound(
                                        fuzzy.norm.Min.Min(),
                                        fuzzy.operator.Input.Input( system.variables["input_Length_MS"].adjectives[PERFECT_SCORE] ),
                                        fuzzy.operator.Input.Input( system.variables["input_JavaEntropy_MS"].adjectives[HIGH_SCORE] ) )
                                    )
    )

    return system

def export_system(system, directory) :
    from fuzzy.doc.plot.gnuplot import doc
    
    d = doc.Doc(directory)
    d.createDoc(system)

    import fuzzy.doc.structure.dot.dot
    import subprocess
    for name,rule in system.rules.items():
            cmd = "dot -T png -o '%s/fuzzy-Rule %s.png'" % (directory,name)
            f = subprocess.Popen(cmd, shell=True, bufsize=32768, stdin=subprocess.PIPE).stdin
            fuzzy.doc.structure.dot.dot.print_header(f,"XXX")
            fuzzy.doc.structure.dot.dot.print_dot(rule,f,system,"")
            fuzzy.doc.structure.dot.dot.print_footer(f)
    cmd = "dot -T png -o '%s/fuzzy-System.png'" % directory
    f = subprocess.Popen(cmd, shell=True, bufsize=32768, stdin=subprocess.PIPE).stdin
    fuzzy.doc.structure.dot.dot.printDot(system,f)

    d.overscan=0
    in_vars = [name for name,var in system.variables.items() if isinstance(var,fuzzy.InputVariable.InputVariable)]
    out_vars = [name for name,var in system.variables.items() if isinstance(var,fuzzy.OutputVariable.OutputVariable)]
    
    if len(in_vars) == 2 and not (
            isinstance(system.variables[in_vars[0]].fuzzify,fuzzy.fuzzify.Dict.Dict)
        or
            isinstance(system.variables[in_vars[1]].fuzzify,fuzzy.fuzzify.Dict.Dict)
    ):
        for out_var in out_vars:
            args = []
            if isinstance(system.variables[out_var].defuzzify,fuzzy.defuzzify.Dict.Dict):
                for adj in system.variables[out_var].adjectives:
                    d.create3DPlot_adjective(system, in_vars[0], in_vars[1], out_var, adj, {})
            else:
                d.create3DPlot(system, in_vars[0], in_vars[1], out_var, {})

class RiskIndicator :
    def __init__(self) :
        self.risk_analysis_obj = []

    def add_risk_analysis(self, obj) :
        self.risk_analysis_obj.append( obj )

    def with_apk(self, apk_file) :
      if apk_file.is_valid_APK() :
        d = dvm.DalvikVMFormat( apk_file.get_dex() )
        dx = analysis.uVMAnalysis( d )

        return self.with_apk_direct(apk_file, d, dx)
      return {}
 
    def with_apk_direct(self, apk_file, d, dx) :
      res = {}
      for i in self.risk_analysis_obj :
        res[ i.get_name() ] = i.with_apk( apk_file, d, dx )
      return res

    def with_dex(self, dex_file) :
      """
            @param dex_file : a buffer

            @rtype : return the risk of the dex file (from 0.0 to 100.0)
      """
      d = dvm.DalvikVMFormat( dex_file )
      dx = analysis.uVMAnalysis( d )
      
      return self.with_dex_direct(d, dx)

    def with_dex_direct(self, d, dx) :
      res = {}
      for i in self.risk_analysis_obj :
        res[ i.get_name() ] = i.with_dex( d, dx )
      return res

class FuzzyRisk :
  """
  Calculate the risk to install a specific android application by using :
        Permissions : 
            - dangerous
            - signatureOrSystem
            - signature
            - normal
        
            - money
            - internet
            - sms
            - call
            - privacy

         API :
            - DexClassLoader

         Files :
            - binary file
            - shared library

        note : pyfuzzy without fcl support (don't install antlr)
  """
  def __init__(self) :
    self.system = create_system_risk()
#     export_system( SYSTEM, "./output" )
        
    self.system_method_risk = create_system_method_one_score()

  def get_name(self) :
    return "FuzzyRisk"

  def with_apk(self, apk_file, d, dx) :
    """
      @param apk_file : an L{APK} object

      @rtype : return the risk of the apk file (from 0.0 to 100.0)
    """
    risks = { DANGEROUS_RISK    : 0.0,
              MONEY_RISK        : 0.0,
              PRIVACY_RISK      : 0.0,
              INTERNET_RISK     : 0.0,
              BINARY_RISK       : 0.0,
              DYNAMIC_RISK      : 0.0,
            }

    self.__eval_risk_perm( apk_file.get_details_permissions(), risks )

    self.__eval_risk_dyn( dx, risks )
    self.__eval_risk_bin( apk_file.get_files_types(), risks )

    val = self.__eval_risks( risks )

    return val

  def with_dex(self, vm, vmx) :
      risks = { DANGEROUS_RISK    : 0.0,
                MONEY_RISK        : 0.0,
                PRIVACY_RISK      : 0.0,
                INTERNET_RISK     : 0.0,
                BINARY_RISK       : 0.0,
                DYNAMIC_RISK      : 0.0,
              }
      

      d = {}
      for i in vmx.get_permissions( [] ) :
          d[ i ] = DVM_PERMISSIONS["MANIFEST_PERMISSION"][i] 
      self.__eval_risk_perm( d, risks )
      self.__eval_risk_dyn( vmx, risks )
      
      val = self.__eval_risks( risks )
      
      return val

  def test(self) :
      ##########################
      score_order_sign = {}


      import sys
      sys.path.append("./elsim")
      from elsim.elsign.libelsign import libelsign
      for method in vm.get_methods() :
          if method.get_length() < 80 :
              continue

          score_order_sign[ method ] = self.get_method_score( method.get_length(),
                               libelsign.entropy( vmx.get_method_signature(method, "L4", { "L4" : { "arguments" : ["Landroid"] } } ).get_string() ),
                               libelsign.entropy( vmx.get_method_signature(method, "L4", { "L4" : { "arguments" : ["Ljava"] } } ).get_string() ),
                               map(lambda perm : (perm, DVM_PERMISSIONS["MANIFEST_PERMISSION"][ perm ]), vmx.get_permissions_method( method )),
          )

          
      for v in sorted(score_order_sign, key=lambda x : score_order_sign[x], reverse=True) :
          print v.get_name(), v.get_class_name(), v.get_descriptor(), v.get_length(), score_order_sign[ v ]

      ##########################

      return val, score_order_sign

  def __eval_risk_perm(self, list_details_permissions, risks) :
      for i in list_details_permissions :
          permission = i
          if permission.find(".") != -1 :
              permission = permission.split(".")[-1]
#            print permission, GENERAL_PERMISSIONS_RISK[ list_details_permissions[ i ][0] ]
        
          risk_type = GENERAL_PERMISSIONS_RISK[ list_details_permissions[ i ][0] ]

          risks[ DANGEROUS_RISK ] += RISK_VALUES [ risk_type ]

          try :
              for j in PERMISSIONS_RISK[ permission ] :
                  risks[ j ] += RISK_VALUES[ j ]
          except KeyError :
              pass

  def __eval_risk_dyn(self, vmx, risks) :
      for m, _ in vmx.tainted_packages.get_packages() :
          if m.get_name() == "Ldalvik/system/DexClassLoader;" :
              for path in m.get_paths() :
                  if path.get_access_flag() == analysis.TAINTED_PACKAGE_CREATE :
                      risks[ DYNAMIC_RISK ] = RISK_VALUES[ DYNAMIC_RISK ]
                      return

  def __eval_risk_bin(self, list_details_files, risks) :
      for i in list_details_files :
          if "ELF" in list_details_files[ i ] :
              # shared library
              if "shared" in list_details_files[ i ] :
                  risks[ BINARY_RISK ] += RISK_VALUES [ BINARY_RISK ]
              # binary 
              else :
                  risks[ BINARY_RISK ] += RISK_VALUES [ EXPLOIT_RISK ]

  def __eval_risks(self, risks) :
      output_values = {"output_malware_risk" : 0.0}
      input_val = {}
      input_val['input_Dangerous_Risk'] = risks[ DANGEROUS_RISK ]
      input_val['input_Money_Risk'] = risks[ MONEY_RISK ]
      input_val['input_Privacy_Risk'] = risks[ PRIVACY_RISK ]
      input_val['input_Binary_Risk'] = risks[ BINARY_RISK ]
      input_val['input_Internet_Risk'] = risks[ INTERNET_RISK ]
      input_val['input_Dynamic_Risk'] = risks[ DYNAMIC_RISK ]

      #print input_val,
      
      self.system.calculate(input=input_val, output = output_values)

      val = output_values[ "output_malware_risk" ]
      return { "VALUE" : val }
 
  def get_method_score(self, length, android_entropy, java_entropy, permissions) :
      val_permissions = 0
      for i in permissions :
          val_permissions += RISK_VALUES[ GENERAL_PERMISSIONS_RISK[ i[1][0] ] ]

          try :
              for j in PERMISSIONS_RISK[ i[0] ] :
                  val_permissions += RISK_VALUES[ j ]
          except KeyError :
              pass
      
      print length, android_entropy, java_entropy, val_permissions

      output_values = {"output_method_one_score" : 0.0}
      input_val = {}
      input_val['input_Length_MS'] = length
      input_val['input_AndroidEntropy_MS'] = android_entropy
      input_val['input_JavaEntropy_MS'] = java_entropy
      input_val['input_Permissions_MS'] = val_permissions
     
      self.system_method_risk.calculate(input=input_val, output = output_values)
      score = output_values[ "output_method_one_score" ]

      return score

  def simulate(self, risks) :
      return self.__eval_risks( risks )

class RedFlags :
        # APK
          # BINARY 
            # shared library
            # executable
            # dex
            # apk
            # jar
            # shell script
        # Perm
            # SMS
            # Call
            # Money
            # Internet
            # Privacy
            # Normal
            # Dangerous
            # Signature
            # System
        # DEX
            # Obfuscation
    def __init__(self) :
        self.flags = { "APK" : {
                      "SHARED LIBRARIES" :    0,      # presence of shared libraries (ELF)
                      "EXECUTABLE" :          0,      # presence of executables (ELF)
                      "DEX" :                 0,      # presence of dex files 
                      "APK" :                 0,      # presence of APK files
                      "ZIP" :                 0,      # presence of zip files
                      "SHELL_SCRIPT" :        0,      # presence of shell scripts
                    },
                "PERM" : {
                      "SMS" :                 0,      # presence of permissions which can manipulate sms
                      "CALL" :                0,      # presence of permissions which can perform a call
                      "GPS" :                 0,      # presence of permissions which can manipulate your location
                      "MONEY" :               0,      # presence of permissions which can result to a payement
                      "INTERNET" :            0,      # presence of permissions which can access to internet
                      "PRIVACY" :             0,      # presence of permissions which can access to private information
                      "NORMAL" :              0,      # "The default value. A lower-risk permission that gives requesting applications access to isolated application-level features, with minimal risk to other applications, the system, or the user"
                      "DANGEROUS" :           0,      # "A higher-risk permission that would give a requesting application access to private user data or control over the device that can negatively impact the user"
                      "SIGNATURE" :           0,      # "A permission that the system grants only if the requesting application is signed with the same certificate as the application that declared the permission"
                      "SIGNATUREORSYSTEM" :   0,      # "A permission that the system grants only to applications that are in the Android system image or that are signed with the same certificates as those in the system image"
                    },
                "DEX" : {
                      "REFLECTION" :          0,      # presence of the reflection API
                      "NATIVE" :              0,      # presence of loading a shared library
                      "DYNAMIC" :             0,      # presence of loading dynamically a new dex file
                      "CRYPTO" :              0,      # presence of crypto functions
                    },
                #"OBFUSCATION" : {                       # presence of obfuscation techniques
                #}
        }

        self.flags_dex = { "DEX" : self.flags["DEX"] }

    def get_name(self) :
        return "RedFlags"

    def with_apk(self, apk_file, d, dx) :
      flags = copy.deepcopy( self.flags )

      self.analyze_apk( apk_file, flags["APK"] )
      self.analyze_axml( apk_file, flags["PERM"] )
      self.analyze_dex( d, dx, flags["DEX"] )

      return flags

    def with_dex(self, d, dx) :
      flags = self.flags_dex.copy()

      self.analyze_dex( d, dx, flags["DEX"] )
      
      return flags

    def analyze_apk(self, a, flags) :
      elf_executable = [ re.compile("ELF.+executable.+"), "EXECUTABLE" ]
      lib_elf = [ re.compile("ELF.+shared object"), "SHARED LIBRARIES" ]
      apk_file = [ re.compile("Android application package file"), "APK" ]
      dex_file = [ re.compile("Dalvik dex file version 035"), "DEX", re.compile("^classes.dex$") ]
      script_file = [ re.compile("script text executable"), "SHELL_SCRIPT" ]
      zip_file = [ re.compile("^Zip archive data.+"), "ZIP" ]

      regexp = [ elf_executable, lib_elf, apk_file, dex_file, script_file, zip_file ]

      files_types = a.get_files_types()
      for i in files_types :
        for j in regexp :
          if j[0].search( files_types[i] ) != None :
            if len(j) < 3 :
              flags[j[1]] += 1
            else :
              if j[2].search( i ) == None :
                flags[j[1]] += 1

    def analyze_axml(self, a, flags) :
      perms = {
                "SEND_SMS"                  : [ "MONEY",    "SMS" ],
                "SEND_SMS_NO_CONFIRMATION"  : [ "MONEY",    "SMS"],
                "READ_SMS"                  : [ "SMS",      "PRIVACY" ],
                "WRITE_SMS"                 : [ "MONEY",    "SMS" ],
                "RECEIVE_SMS"               : [ "SMS",      "PRIVACY" ],
                "RECEIVE_MMS"               : [ "SMS",      "PRIVACY" ],

                "PHONE_CALL"                : [ "MONEY",    "CALL" ],
                "PROCESS_OUTGOING_CALLS"    : [ "MONEY",    "CALL" ],
                "CALL_PRIVILEGED"           : [ "MONEY",    "CALL" ],

                "INTERNET"                  : [ "INTERNET" ],

                "READ_PHONE_STATE"          : [ "PRIVACY" ],
                
                "READ_CONTACTS"             : [ "PRIVACY" ],
                "WRITE_CONTACTS"            : [ "PRIVACY" ],
                
                "READ_HISTORY_BOOKMARKS"    : [ "PRIVACY" ],
                "WRITE_HISTORY_BOOKMARKS"   : [ "PRIVACY" ],
                
                "READ_PROFILE"              : [ "PRIVACY" ],
                "WRITE_PROFILE"             : [ "PRIVACY" ],
                
                "READ_SOCIAL_STREAM"        : [ "PRIVACY" ],
                "WRITE_SOCIAL_STREAM"       : [ "PRIVACY" ],
                
                "READ_CALENDAR"             : [ "PRIVACY" ],
                "WRITE_CALENDAR"            : [ "PRIVACY" ],

                "READ_USER_DICTIONARY"      : [ "PRIVACY" ],
                "WRITE_USER_DICTIONARY"     : [ "PRIVACY" ],

                "SET_ALARM"                 : [ "PRIVACY" ],

                "ADD_VOICEMAIL"             : [ "PRIVACY" ],

                "GET_ACCOUNTS"              : [ "PRIVACY" ],
                "MANAGE_ACCOUNTS"           : [ "PRIVACY" ],

                "RECORD_AUDIO"              : [ "PRIVACY" ],
                "CAMERA"                    : [ "PRIVACY" ],
                

                "ACCESS_FINE_LOCATION"      : [ "PRIVACY",  "GPS" ],
                "ACCESS_COARSE_LOCATION"    : [ "PRIVACY",  "GPS" ],
                "ACCESS_LOCATION_EXTRA_COMMANS" : [ "GPS"],
                "INSTALL_LOCATION_PROVIDER" : [ "GPS" ],
      }

      for i in a.get_permissions() :
        perm = i.split(".")[-1]

        try :
          flags[ DVM_PERMISSIONS["MANIFEST_PERMISSION"][perm][0].upper() ] += 1

          for j in perms :
            if j == perm :
              for k in perms[j] :
                flags[k] += 1
        except :
          debug("Unknown permission %s" % perm)

    def analyze_dex(self, d, dx, flags) :
      flags["REFLECTION"] = int( analysis.is_reflection_code(dx) )
      flags["NATIVE"] = int( analysis.is_native_code(dx) )
      flags["DYNAMIC"] = int( analysis.is_dyn_code(dx) )
      flags["CRYPTO"] = int( analysis.is_crypto_code(dx) )

class MethodScore :
    def __init__(self, length, matches, android_entropy, java_entropy, permissions, similarity_matches) :
        self.system = create_system_method_score()
        #export_system( self.system, "./output" )

        
        val_permissions = 0
        for i in permissions :
            val_permissions += RISK_VALUES[ GENERAL_PERMISSIONS_RISK[ i[1][0] ] ]

            try :
                for j in PERMISSIONS_RISK[ i[0] ] :
                    val_permissions += RISK_VALUES[ j ]
            except KeyError :
                pass
        
        print length, matches, android_entropy, java_entropy, similarity_matches, val_permissions

        output_values = {"output_method_score" : 0.0}
        input_val = {}
        input_val['input_Length_MS'] = length
        input_val['input_Match_MS'] = matches
        input_val['input_AndroidEntropy_MS'] = android_entropy
        input_val['input_JavaEntropy_MS'] = java_entropy
        input_val['input_Permissions_MS'] = val_permissions
        input_val['input_Similarity_MS'] = similarity_matches
       
        self.system.calculate(input=input_val, output = output_values)
        self.score = output_values[ "output_method_score" ]

    def get_score(self) :
        return self.score


# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

from subprocess import Popen, PIPE, STDOUT

import os, sys
import xmlrpclib

import cPickle

class _Method :
    def __init__(self, proxy, name) :
        self.proxy = proxy
        self.name = name
    
    def __call__(self, *args):
        #print "CALL", self.name, args
        z = getattr( self.proxy, self.name, None )
        #print "SEND", repr(cPickle.dumps( args ) )
       
        try :
            if len(args) == 1 :
                ret = z( cPickle.dumps( args[0] ) )
            else :
                ret = z( cPickle.dumps( args ) )
            #print "RECEIVE", repr(ret)
            return cPickle.loads( ret )
        except xmlrpclib.ProtocolError :
            return []

class MyXMLRPC :
    def __init__(self, proxy) :
        self.proxy = proxy

    def __getattr__(self, name) :
        return _Method(self.proxy, name)

class BasicBlock :
    def __init__(self, ins) :
        self.ins = ins

    def show(self) :
        for i in self.ins :
            print i

class Function :
    def __init__(self, name, start_ea, instructions, information) :
        #print name, start_ea
                
        self.name = name
        self.start_ea = start_ea
        self.information = information
        self.basic_blocks = []
        self.instructions = instructions

        r = {}
        idx = 0
        for i in instructions :
            r[ i[0] ] = idx
            idx += 1

        for i in information[0] :
            try :
                start = r[i[0]]
                end = r[i[1]] + 1
                self.basic_blocks.append( BasicBlock( instructions[start:end] ) )
            except KeyError :
                pass

    def get_instructions(self) :
        return [ i for i in self.instructions ]

def run_ida(idapath, wrapper_init_path, binpath) :
    os.environ["TVHEADLESS"] = "1"
    pid = os.fork()
    if pid == 0:
        wrapper_path = "-S" + wrapper_init_path
        l = [ idapath, "-A", wrapper_path, binpath ]
        print l
        compile = Popen(l, stdout=open('/dev/null', 'w'), stderr=STDOUT)
        stdout, stderr = compile.communicate()
#        print stdout, stderr
        sys.exit(0)

class IDAPipe :
    def __init__(self, idapath, binpath, wrapper_init_path) :
        self.idapath = idapath
        self.binpath = binpath

        self.proxy = None
        
        run_ida(self.idapath, self.binpath, wrapper_init_path)

        while 1 :
            try :
                self.proxy = xmlrpclib.ServerProxy("http://localhost:9000/")
                self.proxy.is_connected()
                break
            except :
                pass
            
        #print self.proxy
        self.proxy = MyXMLRPC( self.proxy )

    def quit(self) :
        try :
            self.proxy.quit()
        except :
            pass

    def _build_functions(self, functions) :
        F = {}

        for i in functions :
            F[ i ] = Function( functions[i][0], i, functions[i][1:-1], functions[i][-1] )

        return F

    def get_quick_functions(self) :
        functions = self.get_raw()
        return self._build_functions( functions )

    def get_raw(self) :
        return self.proxy.get_raw()

    def get_nb_functions(self) :
        return len(self.proxy.Functions())

    def get_functions(self) :
        for function_ea in self.proxy.Functions() :
            self.get_function_addr( function_ea )

    def get_function_name(self, name) :
        function_ea = self.proxy.get_function( name )
        self.get_function_addr( function_ea )

    def get_function_addr(self, function_ea) :
        if function_ea == -1 :
            return

        f_start = function_ea 
        f_end = self.proxy.GetFunctionAttr(function_ea, 4) #FUNCATTR_END)
        
        edges = set()
        boundaries = set((f_start,))
        
        for head in self.proxy.Heads(f_start, f_end) :
            if self.proxy.isCode( self.proxy.GetFlags( head ) ) :
                refs = self.proxy.CodeRefsFrom(head, 0)
                refs = set(filter(lambda x: x>=f_start and x<=f_end, refs))

                #print head, f_end, refs, self.proxy.GetMnem(head), self.proxy.GetOpnd(head, 0), self.proxy.GetOpnd(head, 1)

                if refs :
                    next_head = self.proxy.NextHead(head, f_end)
                    if self.proxy.isFlow(self.proxy.GetFlags(next_head)):
                        refs.add(next_head)
                                                      
                    # Update the boundaries found so far.
                    boundaries.update(refs)
                                                                                                  
                    # For each of the references found, and edge is
                    # created.
                    for r in refs:
                    # If the flow could also come from the address
                    # previous to the destination of the branching
                    # an edge is created.
                        if self.proxy.isFlow(self.proxy.GetFlags(r)):
                            edges.add((self.proxy.PrevHead(r, f_start), r))
                        edges.add((head, r))
                

        #print edges, boundaries
        # Let's build the list of (startEA, startEA) couples
        # for each basic block
        sorted_boundaries = sorted(boundaries, reverse = True)
        end_addr = self.proxy.PrevHead(f_end, f_start)
        bb_addr = []
        for begin_addr in sorted_boundaries:
            bb_addr.append((begin_addr, end_addr))
            # search the next end_addr which could be
            # farther than just the previous head
            # if data are interlaced in the code
            # WARNING: it assumes it won't epicly fail ;)
            end_addr = self.proxy.PrevHead(begin_addr, f_start)
            while not self.proxy.isCode(self.proxy.GetFlags(end_addr)):
                end_addr = self.proxy.PrevHead(end_addr, f_start)
        # And finally return the result
        bb_addr.reverse()
        #print bb_addr, sorted(edges)

def display_function(f) :
    print f, f.name, f.information

    for i in f.basic_blocks :
        print i
        i.show()



# This file is part of Androguard.
#
# Copyright (c) 2012 Geoffroy Gueguen <geoffroy.gueguen@gmail.com>
# All Rights Reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.


class IRForm(object):
    def __init__(self):
        self.var_map = {}

    def is_call(self):
        return False

    def is_cond(self):
        return False

    def is_const(self):
        return False

    def is_ident(self):
        return False

    def is_propagable(self):
        return True

    def get_type(self):
        return None

    def has_side_effect(self):
        return False

    def get_used_vars(self):
        return []

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            from dad.util import log
            log('modify_rhs not implemented %s' % self, 'error')

    def remove_defined_var(self):
        pass

    def get_rhs(self):
        return []

    def get_lhs(self):
        return None

    def visit(self, visitor):
        pass


class Constant(IRForm):
    def __init__(self, value, atype, int_value=None):
        self.v = 'c%s' % value
        self.cst = value
        if int_value is None:
            self.cst2 = value
        else:
            self.cst2 = int_value
        self.type = atype

    def get_used_vars(self):
        return []

    def is_call(self):
        return False

    def is_const(self):
        return True

    def has_side_effect(self):
        return False

    def visit(self, visitor, to_int=False):
        if self.type == 'Z':
            if self.cst == 0:
                return visitor.visit_constant('false')
            else:
                return visitor.visit_constant('true')
        elif self.type == 'class':
            return visitor.visit_base_class(self.cst)
        elif to_int:
            return visitor.visit_constant(self.cst2)
        else:
            return visitor.visit_constant(self.cst)


class BaseClass(IRForm):
    def __init__(self, name):
        self.v = 'c%s' % name
        self.cls = name

    def is_const(self):
        return True

    def visit(self, visitor):
        return visitor.visit_base_class(self.cls)


class Variable(IRForm):
    def __init__(self, value):
        self.v = value

    def get_type(self):
        return None

    def get_used_vars(self):
        return [self.v]

    def is_call(self):
        return False

    def has_side_effect(self):
        return False

    def is_ident(self):
        return True

    def visit(self, visitor):
        return visitor.visit_variable(self.v)

    def visit_decl(self, visitor):
        return visitor.visit_decl(self.v)


class Param(IRForm):
    def __init__(self, value, atype):
        self.v = value
        self.type = atype

    def get_type(self):
        return self.type

    def get_used_vars(self):
        return [self.v]

    def is_call(self):
        return False

    def has_side_effect(self):
        return False

    def is_ident(self):
        return True

    def visit(self, visitor):
        return visitor.visit_param(self.v)

    def visit_decl(self, visitor):
        pass


class ThisParam(Param):
    def __init__(self, value, atype):
        super(ThisParam, self).__init__(value, atype)

    def is_const(self):
        return True

    def get_used_vars(self):
        return []

    def is_ident(self):
        return True

    def visit(self, visitor):
        return visitor.visit_this()

    def visit_decl(self, visitor):
        pass


class AssignExpression(IRForm):
    def __init__(self, lhs, rhs):
        super(AssignExpression, self).__init__()
        self.lhs = lhs.v
        self.rhs = rhs
        self.var_map[lhs.v] = lhs

    def is_propagable(self):
        return self.rhs.is_propagable()

    def is_call(self):
        return self.rhs.is_call()

    def has_side_effect(self):
        return self.rhs.has_side_effect()

    def get_rhs(self):
        return self.rhs

    def get_lhs(self):
        return self.lhs

    def get_used_vars(self):
        return self.rhs.get_used_vars()

    def remove_defined_var(self):
        self.lhs = None

    def modify_rhs(self, old, new):
        self.rhs.modify_rhs(old, new)

    def visit(self, visitor):
        return visitor.visit_assign(self.var_map.get(self.lhs), self.rhs)


class MoveResultExpression(IRForm):
    def __init__(self, lhs, rhs):
        super(MoveResultExpression, self).__init__()
        self.lhs = lhs.v
        self.rhs = rhs.v
        self.var_map.update([(lhs.v, lhs), (rhs.v, rhs)])

    def is_propagable(self):
        return self.var_map[self.rhs].is_propagable()

    def is_call(self):
        return self.var_map[self.rhs].is_call()

    def has_side_effect(self):
        return self.var_map[self.rhs].has_side_effect()

    def get_used_vars(self):
        return self.var_map[self.rhs].get_used_vars()

    def get_rhs(self):
        return self.var_map[self.rhs]

    def get_lhs(self):
        return self.lhs

    def visit(self, visitor):
        m = self.var_map
        return visitor.visit_move_result(m[self.lhs], m[self.rhs])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            self.var_map[self.rhs].modify_rhs(old, new)


class MoveExpression(IRForm):
    def __init__(self, lhs, rhs):
        super(MoveExpression, self).__init__()
        self.lhs = lhs.v
        self.rhs = rhs.v
        self.var_map.update([(lhs.v, lhs), (rhs.v, rhs)])

    def has_side_effect(self):
        return False

    def is_call(self):
        return self.var_map[self.rhs].is_call()

    def get_used_vars(self):
        return self.var_map[self.rhs].get_used_vars()

    def get_rhs(self):
        return self.var_map[self.rhs]

    def get_lhs(self):
        return self.lhs

    def visit(self, visitor):
        m = self.var_map
        return visitor.visit_move(m[self.lhs], m[self.rhs])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            rhs = self.var_map[self.rhs]
            if not (rhs.is_const() or rhs.is_ident()):
                rhs.modify_rhs(old, new)


class ArrayStoreInstruction(IRForm):
    def __init__(self, rhs, array, index):
        super(ArrayStoreInstruction, self).__init__()
        self.rhs = rhs.v
        self.array = array.v
        self.index = index.v
        self.var_map.update([(rhs.v, rhs), (array.v, array), (index.v, index)])

    def has_side_effect(self):
        return True

    def get_used_vars(self):
        m = self.var_map
        lused_vars = m[self.array].get_used_vars()
        lused_vars.extend(m[self.index].get_used_vars())
        lused_vars.extend(m[self.rhs].get_used_vars())
        return lused_vars

    def visit(self, visitor):
        m = self.var_map
        return visitor.visit_astore(m[self.array], m[self.index], m[self.rhs])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arr = self.var_map[self.array]
            idx = self.var_map[self.index]
            rhs = self.var_map[self.rhs]
            for arg in (arr, idx, rhs):
                if not (arg.is_const() or arg.is_ident()):
                    arg.modify_rhs(old, new)


class StaticInstruction(IRForm):
    def __init__(self, rhs, klass, ftype, name):
        super(StaticInstruction, self).__init__()
        self.rhs = rhs.v
        self.cls = klass
        self.ftype = ftype
        self.name = name
        self.var_map[rhs.v] = rhs

    def has_side_effect(self):
        return True

    def get_used_vars(self):
        return self.var_map[self.rhs].get_used_vars()

    def get_lhs(self):
        return None

    def visit(self, visitor):
        m = self.var_map
        return visitor.visit_put_static(self.cls, self.name, m[self.rhs])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            self.var_map[self.rhs].modify_rhs(old, new)


class InstanceInstruction(IRForm):
    def __init__(self, rhs, lhs, klass, atype, name):
        super(InstanceInstruction, self).__init__()
        self.lhs = lhs.v
        self.rhs = rhs.v
        self.atype = atype
        self.cls = klass
        self.name = name
        self.var_map.update([(lhs.v, lhs), (rhs.v, rhs)])

    def has_side_effect(self):
        return True

    def get_used_vars(self):
        m = self.var_map
        lused_vars = m[self.lhs].get_used_vars()
        lused_vars.extend(m[self.rhs].get_used_vars())
        return lused_vars

    def get_lhs(self):
        return None

    def visit(self, visitor):
        m = self.var_map
        return visitor.visit_put_instance(m[self.lhs], self.name, m[self.rhs])

    def modify_rhs(self, old, new):
        v_m = self.var_map
        if old in v_m:
            v_m[old] = new
        else:
            lhs = v_m[self.lhs]
            rhs = v_m[self.rhs]
            for arg in (lhs, rhs):
                if not (arg.is_const() or arg.is_ident()):
                    arg.modify_rhs(old, new)


class NewInstance(IRForm):
    def __init__(self, ins_type):
        super(NewInstance, self).__init__()
        self.type = ins_type

    def get_type(self):
        return self.type

#    def has_side_effect(self):
#        return True

    def get_used_vars(self):
        return []

    def visit(self, visitor):
        return visitor.visit_new(self.type)

    def modify_rhs(self, old, new):
        pass


class InvokeInstruction(IRForm):
    def __init__(self, clsname, name, base, rtype, ptype, nbargs, args):
        super(InvokeInstruction, self).__init__()
        self.cls = clsname
        self.name = name
        self.base = base.v
        self.rtype = rtype
        self.ptype = ptype
        self.nbargs = nbargs
        self.args = [arg.v for arg in args]
        self.var_map[base.v] = base
        for arg in args:
            self.var_map[arg.v] = arg

    def get_type(self):
        return self.rtype

    def is_call(self):
        return True

    def has_side_effect(self):
        return True

    def modify_rhs(self, old, new):
        if old in self.var_map:
            arg = self.var_map[old]
            if not (arg.is_ident() or arg.is_const()):
                arg.modify_rhs(old, new)
            else:
                self.var_map[old] = new
        else:
            base = self.var_map[self.base]
            if not (base.is_ident() or base.is_const()):
                base.modify_rhs(old, new)
            for arg in self.args:
                cnt = self.var_map[arg]
                if cnt.is_ident() or cnt.is_const():
                    continue
                else:
                    self.var_map[arg].modify_rhs(old, new)

    def get_used_vars(self):
        m = self.var_map
        lused_vars = []
        for a in self.args:
            lused_vars.extend(m[a].get_used_vars())
        lused_vars.extend(m[self.base].get_used_vars())
        return lused_vars

    def visit(self, visitor):
        m = self.var_map
        largs = [m[arg] for arg in self.args]
        return visitor.visit_invoke(self.name, m[self.base], largs)


class InvokeRangeInstruction(InvokeInstruction):
    def __init__(self, clsname, name, rtype, ptype, nbargs, args):
        base = args.pop(0)
        super(InvokeRangeInstruction, self).__init__(clsname, name, base,
                                                    rtype, ptype, nbargs, args)


class InvokeDirectInstruction(InvokeInstruction):
    def __init__(self, clsname, name, base, rtype, ptype, nbargs, args):
        super(InvokeDirectInstruction, self).__init__(clsname, name, base,
                                                    rtype, ptype, nbargs, args)


class InvokeStaticInstruction(InvokeInstruction):
    def __init__(self, clsname, name, base, rtype, ptype, nbargs, args):
        # TODO: check base class name and current class name
        super(InvokeStaticInstruction, self).__init__(clsname, name, base,
                                                    rtype, ptype, nbargs, args)

    def get_used_vars(self):
        return list(set(self.args))


class ReturnInstruction(IRForm):
    def __init__(self, arg):
        super(ReturnInstruction, self).__init__()
        self.arg = arg
        if arg is not None:
            self.var_map[arg.v] = arg
            self.arg = arg.v

    def get_used_vars(self):
        if self.arg is None:
            return []
        return self.var_map[self.arg].get_used_vars()

    def get_lhs(self):
        return None

    def visit(self, visitor):
        if self.arg is None:
            return visitor.visit_return_void()
        else:
            return visitor.visit_return(self.var_map[self.arg])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arg = self.var_map[self.arg]
            if not (arg.is_const() or arg.is_ident()):
                arg.modify_rhs(old, new)


class NopExpression(IRForm):
    def __init__(self):
        pass

    def get_used_vars(self):
        return []

    def get_lhs(self):
        return None

    def visit(self, visitor):
        return visitor.visit_nop()


class SwitchExpression(IRForm):
    def __init__(self, src, branch):
        super(SwitchExpression, self).__init__()
        self.src = src.v
        self.var_map[src.v] = src

    def get_used_vars(self):
        m = self.var_map
        return m[self.src].get_used_vars()

    def visit(self, visitor):
        return visitor.visit_switch(self.var_map[self.src])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            src = self.var_map[self.src]
            if not (src.is_const() or src.is_ident()):
                src.modify_rhs(old, new)


class CheckCastExpression(IRForm):
    def __init__(self, arg, _type):
        super(CheckCastExpression, self).__init__()
        self.arg = arg.v
        self.var_map[arg.v] = arg
        self.type = _type

    def get_used_vars(self):
        return self.var_map[self.arg].get_used_vars()

    def visit(self, visitor):
        return visitor.visit_check_cast(self.var_map[self.arg], self.type)

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arg = self.var_map[self.arg]
            if not (arg.is_const() or arg.is_ident()):
                arg.modify_rhs(old, new)


class ArrayExpression(IRForm):
    def __init__(self):
        super(ArrayExpression, self).__init__()


class ArrayLoadExpression(ArrayExpression):
    def __init__(self, arg, index):
        super(ArrayLoadExpression, self).__init__()
        self.array = arg.v
        self.idx = index.v
        self.var_map.update([(arg.v, arg), (index.v, index)])

    def get_used_vars(self):
        m = self.var_map
        lused_vars = m[self.array].get_used_vars()
        lused_vars.extend(m[self.idx].get_used_vars())
        return lused_vars

    def visit(self, visitor):
        m = self.var_map
        return visitor.visit_aload(m[self.array], m[self.idx])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arr = self.var_map[self.array]
            idx = self.var_map[self.idx]
            for arg in (arr, idx):
                if not (arg.is_const() or arg.is_ident()):
                    arg.modify_rhs(old, new)


class ArrayLengthExpression(ArrayExpression):
    def __init__(self, array):
        super(ArrayLengthExpression, self).__init__()
        self.array = array.v
        self.var_map[array.v] = array

    def get_type(self):
        return 'I'

    def get_used_vars(self):
        m = self.var_map
        return m[self.array].get_used_vars()

    def visit(self, visitor):
        return visitor.visit_alength(self.var_map[self.array])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arr = self.var_map[self.array]
            if not (arr.is_const() or arr.is_ident()):
                arr.modify_rhs(old, new)


class NewArrayExpression(ArrayExpression):
    def __init__(self, asize, atype):
        super(NewArrayExpression, self).__init__()
        self.size = asize.v
        self.type = atype
        self.var_map[asize.v] = asize

    def is_propagable(self):
        return False

    def get_used_vars(self):
        m = self.var_map
        return m[self.size].get_used_vars()

    def visit(self, visitor):
        return visitor.visit_new_array(self.type, self.var_map[self.size])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            asize = self.var_map[self.size]
            if not (asize.is_const() or asize.is_ident()):
                asize.modify_rhs(old, new)


class FilledArrayExpression(ArrayExpression):
    def __init__(self, asize, atype, args):
        super(FilledArrayExpression, self).__init__()
        self.size = asize.v
        self.var_map[asize.v] = asize
        self.type = atype
        self.args = []
        for arg in args:
            self.var_map[arg.v] = arg
            self.args.append(arg.v)

    def is_propagable(self):
        return False

    def has_side_effect(self):
        return True

    def get_used_vars(self):
        lused_vars = []
        for arg in self.args:
            lused_vars.extend(self.var_map[arg].get_used_vars())
        lused_vars.append(self.size)
        return lused_vars

    def visit(self, visitor):
        m = self.var_map
        largs = [m[arg] for arg in self.args]
        return visitor.visit_filled_new_array(self.type, m[self.size], largs)


class FillArrayExpression(ArrayExpression):
    def __init__(self, reg, value):
        super(FillArrayExpression, self).__init__()
        self.reg = reg.v
        self.var_map[reg.v] = reg
        self.value = value

    def is_propagable(self):
        return False

    def get_rhs(self):
        return self.reg

    def get_used_vars(self):
        m = self.var_map
        return m[self.reg].get_used_vars()

    def visit(self, visitor):
        return visitor.visit_fill_array(self.var_map[self.reg], self.value)


class RefExpression(IRForm):
    def __init__(self, ref):
        super(RefExpression, self).__init__()
        self.ref = ref.v
        self.var_map[ref.v] = ref

    def get_used_vars(self):
        m = self.var_map
        return m[self.ref].get_used_vars()

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            self.var_map[self.ref].modify_rhs(old, new)


class MonitorEnterExpression(RefExpression):
    def __init__(self, ref):
        super(MonitorEnterExpression, self).__init__(ref)

    def visit(self, visitor):
        return visitor.visit_monitor_enter(self.var_map[self.ref])


class MonitorExitExpression(RefExpression):
    def __init__(self, ref):
        super(MonitorExitExpression, self).__init__(ref)

    def visit(self, visitor):
        return visitor.visit_monitor_exit(self.var_map[self.ref])


class ThrowExpression(RefExpression):
    def __init__(self, ref):
        super(ThrowExpression, self).__init__(ref)

    def visit(self, visitor):
        return visitor.visit_throw(self.var_map[self.ref])


class BinaryExpression(IRForm):
    def __init__(self, op, arg1, arg2):
        super(BinaryExpression, self).__init__()
        self.op = op
        self.arg1 = arg1.v
        self.arg2 = arg2.v
        self.var_map.update([(arg1.v, arg1), (arg2.v, arg2)])

    # TODO: return the max type of arg1 & arg2
    def get_type(self):
        return None

    def has_side_effect(self):
        m = self.var_map
        res = m[self.arg1].has_side_effect() or m[self.arg2].has_side_effect()
        return res

    def get_used_vars(self):
        m = self.var_map
        lused_vars = m[self.arg1].get_used_vars()
        lused_vars.extend(m[self.arg2].get_used_vars())
        return lused_vars

    def visit(self, visitor):
        m = self.var_map
        return visitor.visit_binary_expression(self.op, m[self.arg1],
                                                        m[self.arg2])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arg1 = self.var_map[self.arg1]
            arg2 = self.var_map[self.arg2]
            for arg in (arg1, arg2):
                if not (arg.is_ident() or arg.is_const()):
                    arg.modify_rhs(old, new)


class BinaryCompExpression(BinaryExpression):
    def __init__(self, op, arg1, arg2):
        super(BinaryCompExpression, self).__init__(op, arg1, arg2)

    def visit(self, visitor):
        m = self.var_map
        return visitor.visit_cond_expression(self.op, m[self.arg1],
                                                      m[self.arg2])


class BinaryExpression2Addr(BinaryExpression):
    def __init__(self, op, dest, arg):
        super(BinaryExpression2Addr, self).__init__(op, dest, arg)


class BinaryExpressionLit(BinaryExpression):
    def __init__(self, op, arg1, arg2):
        super(BinaryExpressionLit, self).__init__(op, arg1, arg2)


class UnaryExpression(IRForm):
    def __init__(self, op, arg):
        super(UnaryExpression, self).__init__()
        self.op = op
        self.arg = arg.v
        self.var_map[arg.v] = arg

    def get_type(self):
        return self.var_map[self.arg].get_type()

    def get_used_vars(self):
        m = self.var_map
        return m[self.arg].get_used_vars()

    def visit(self, visitor):
        return visitor.visit_unary_expression(self.op, self.var_map[self.arg])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arg = self.var_map[self.arg]
            if not (arg.is_const() or arg.is_ident()):
                arg.modify_rhs(old, new)


class CastExpression(UnaryExpression):
    def __init__(self, op, atype, arg):
        super(CastExpression, self).__init__(op, arg)
        self.type = atype

    def get_type(self):
        return self.type

    def get_used_vars(self):
        return self.var_map[self.arg].get_used_vars()

    def visit(self, visitor):
        return visitor.visit_cast(self.op, self.var_map[self.arg])


CONDS = {
    '==': '!=',
    '!=': '==',
    '<': '>=',
    '<=': '>',
    '>=': '<',
    '>': '<=',
}


class ConditionalExpression(IRForm):
    def __init__(self, op, arg1, arg2):
        super(ConditionalExpression, self).__init__()
        self.op = op
        self.arg1 = arg1.v
        self.arg2 = arg2.v
        self.var_map.update([(arg1.v, arg1), (arg2.v, arg2)])

    def get_lhs(self):
        return None

    def is_cond(self):
        return True

    def get_used_vars(self):
        m = self.var_map
        lused_vars = m[self.arg1].get_used_vars()
        lused_vars.extend(m[self.arg2].get_used_vars())
        return lused_vars

    def neg(self):
        self.op = CONDS[self.op]

    def visit(self, visitor):
        m = self.var_map
        return visitor.visit_cond_expression(self.op, m[self.arg1],
                                                      m[self.arg2])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arg1 = self.var_map[self.arg1]
            arg2 = self.var_map[self.arg2]
            for arg in (arg1, arg2):
                if not (arg.is_ident() or arg.is_const()):
                    arg.modify_rhs(old, new)


class ConditionalZExpression(IRForm):
    def __init__(self, op, arg):
        super(ConditionalZExpression, self).__init__()
        self.op = op
        self.arg = arg.v
        self.var_map[arg.v] = arg

    def get_lhs(self):
        return None

    def is_cond(self):
        return True

    def get_used_vars(self):
        m = self.var_map
        return m[self.arg].get_used_vars()

    def neg(self):
        self.op = CONDS[self.op]

    def visit(self, visitor):
        return visitor.visit_condz_expression(self.op, self.var_map[self.arg])

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arg = self.var_map[self.arg]
            if not (arg.is_ident() or arg.is_const()):
                arg.modify_rhs(old, new)


class InstanceExpression(IRForm):
    def __init__(self, arg, klass, ftype, name):
        super(InstanceExpression, self).__init__()
        self.arg = arg.v
        self.cls = klass
        self.ftype = ftype
        self.name = name
        self.var_map[arg.v] = arg

    def get_type(self):
        return self.ftype

    def get_used_vars(self):
        m = self.var_map
        return m[self.arg].get_used_vars()

    def visit(self, visitor):
        return visitor.visit_get_instance(self.var_map[self.arg], self.name)

    def modify_rhs(self, old, new):
        if old in self.var_map:
            self.var_map[old] = new
        else:
            arg = self.var_map[self.arg]
            if not (arg.is_ident() or arg.is_const()):
                arg.modify_rhs(old, new)


class StaticExpression(IRForm):
    def __init__(self, cls_name, field_type, field_name):
        super(StaticExpression, self).__init__()
        self.cls = cls_name
        self.ftype = field_type
        self.name = field_name

    def get_type(self):
        return self.ftype

    def visit(self, visitor):
        return visitor.visit_get_static(self.cls, self.name)

    def modify_rhs(self, old, new):
        pass



#!/usr/bin/env python

# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

import sys, os

from optparse import OptionParser

from androguard.core import androconf
from androguard.core.bytecodes import apk

sys.path.append("./elsim/")
from elsim.elsign import dalvik_elsign


option_0 = { 'name' : ('-i', '--input'), 'help' : 'file : use this filename', 'nargs' : 1 }
option_1 = { 'name' : ('-d', '--directory'), 'help' : 'directory : use this directory', 'nargs' : 1 }
option_2 = { 'name' : ('-b', '--database'), 'help' : 'database : use this database', 'nargs' : 1 }
option_3 = { 'name' : ('-c', '--config'), 'help' : 'use this configuration', 'nargs' : 1 }
option_4 = { 'name' : ('-v', '--verbose'), 'help' : 'display debug information', 'action' : 'count' }

options = [option_0, option_1, option_2, option_3, option_4]

def display(ret, debug) :
    print "---->", ret[0]
    sys.stdout.flush()

def main(options, arguments) :
    if options.database == None or options.config == None :
        return

    s = dalvik_elsign.MSignature( options.database, options.config, options.verbose != None, ps = dalvik_elsign.PublicSignature)

    if options.input != None :
        ret_type = androconf.is_android( options.input ) 
        
        print os.path.basename(options.input), ":",
        sys.stdout.flush()
        if ret_type == "APK" :
            try :
                a = apk.APK( options.input )
                if a.is_valid_APK() :
                    display( s.check_apk( a ), options.verbose )
                else :
                    print "INVALID"
            except Exception, e :
                print "ERROR", e

        elif ret_type == "DEX" :
            display( s.check_dex( open(options.input, "rb").read() ), options.verbose )
    elif options.directory != None :
        for root, dirs, files in os.walk( options.directory, followlinks=True ) :
            if files != [] :
                for f in files :
                    real_filename = root
                    if real_filename[-1] != "/" :
                        real_filename += "/"
                    real_filename += f

                    ret_type = androconf.is_android( real_filename )
                    if ret_type == "APK"  :
                        print os.path.basename( real_filename ), ":",
                        sys.stdout.flush()
                        try :
                            a = apk.APK( real_filename )
                            if a.is_valid_APK() :
                                display( s.check_apk( a ), options.verbose )
                            else :
                                print "INVALID APK"
                        except Exception, e :
                            print "ERROR", e
                    elif ret_type == "DEX" :
                        try :
                            print os.path.basename( real_filename ), ":",
                            sys.stdout.flush()
                            display( s.check_dex( open(real_filename, "rb").read() ), options.verbose )
                        except Exception, e : 
                            print "ERROR", e

    elif options.version != None :
        print "Androsign version %s" % androconf.ANDROGUARD_VERSION

if __name__ == "__main__" :
    parser = OptionParser()
    for option in options :
        param = option['name']
        del option['name']
        parser.add_option(*param, **option)

    options, arguments = parser.parse_args()
    sys.argv[:] = arguments
    main(options, arguments)



#!/usr/bin/env python

# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of  
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

from xml.sax.saxutils import escape, unescape
import sys, hashlib, os
from optparse import OptionParser

from androguard.core.bytecodes import apk, dvm
from androguard.core.analysis import analysis, ganalysis
from androguard.core import androconf

option_0 = { 'name' : ('-i', '--input'), 'help' : 'filename input (dex, apk)', 'nargs' : 1 }
option_1 = { 'name' : ('-o', '--output'), 'help' : 'filename output of the gexf', 'nargs' : 1 }

options = [option_0, option_1]

def main(options, arguments) :
    if options.input != None and options.output != None :
        ret_type = androconf.is_android( options.input )
        
        vm = None
        a = None
        if ret_type == "APK"  :
            a = apk.APK( options.input )
            if a.is_valid_APK() :
                vm = dvm.DalvikVMFormat( a.get_dex() )
            else :
                print "INVALID APK"
        elif ret_type == "DEX" :
            try :
                vm = dvm.DalvikVMFormat( open(options.input, "rb").read() )
            except Exception, e :
                print "INVALID DEX", e

        vmx = analysis.VMAnalysis( vm )
        gvmx = ganalysis.GVMAnalysis( vmx, a )

        b = gvmx.export_to_gexf()
        androconf.save_to_disk( b, options.output )

if __name__ == "__main__" :
   parser = OptionParser()
   for option in options :
	  param = option['name']
	  del option['name']
	  parser.add_option(*param, **option)

	  
   options, arguments = parser.parse_args()
   sys.argv[:] = arguments
   main(options, arguments)	



# This file is part of Androguard.
#
# Copyright (c) 2012 Geoffroy Gueguen <geoffroy.gueguen@gmail.com>
# All Rights Reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.


class LoopType(object):
    def __init__(self):
        self.loop = None

    def pretest(self):
        return self.loop == 0

    def posttest(self):
        return self.loop == 1

    def endless(self):
        return self.loop == 2

    def set_pretest(self):
        self.loop = 0

    def set_posttest(self):
        self.loop = 1

    def set_endless(self):
        self.loop = 2


class Node(object):
    def __init__(self, name):
        self.name = name
        self.num = 0
        self.looptype = LoopType()
        self.interval = None
        self.startloop = False
        self.endloop = False
        self.type = -1
        self.latch = None
        self.if_follow = None
        self.loop_follow = None
        self.switch_follow = None
        self.loop_nodes = []

    def copy_from(self, node):
        self.num = node.num
        self.looptype = node.looptype
        self.interval = node.interval
        self.startloop = node.startloop
        self.endloop = node.endloop
        self.type = node.type
        self.latch = node.latch
        self.if_follow = node.if_follow
        self.loop_follow = node.loop_follow
        self.switch_follow = node.switch_follow
        self.loop_nodes = node.loop_nodes

    def update_attribute_with(self, n_map):
        self.latch = n_map.get(self.latch, self.latch)
        self.if_follow = n_map.get(self.if_follow, self.if_follow)
        self.loop_follow = n_map.get(self.loop_follow, self.loop_follow)
        self.switch_follow = n_map.get(self.switch_follow, self.switch_follow)
        self.loop_nodes = list({n_map.get(n, n) for n in self.loop_nodes})

    def is_cond(self):
        return self.type == 0

    def set_cond(self):
        self.type = 0

    def is_switch(self):
        return self.type == 1

    def set_switch(self):
        self.type = 1

    def is_stmt(self):
        return self.type == 2

    def set_stmt(self):
        self.type = 2

    def is_return(self):
        return self.type == 3

    def set_return(self):
        self.type = 3

    def is_throw(self):
        return self.type == 4

    def set_throw(self):
        self.type = 4

    def set_loop_pretest(self):
        self.looptype.set_pretest()

    def set_loop_posttest(self):
        self.looptype.set_posttest()

    def set_loop_endless(self):
        self.looptype.set_endless()

    def get_head(self):
        return self

    def get_end(self):
        return self

    def set_loop_nodes(self, nodes):
        self.loop_nodes = nodes

    def set_start_loop(self, b=True):
        self.startloop = b

    def set_end_loop(self, b=True):
        self.endloop = b

    def set_if_follow(self, node):
        self.if_follow = node

    def get_if_follow(self):
        return self.if_follow

    def set_loop_follow(self, node):
        self.loop_follow = node

    def get_loop_follow(self):
        return self.loop_follow

    def set_switch_follow(self, node):
        self.switch_follow = node

    def get_switch_follow(self):
        return self.switch_follow

    def set_latch_node(self, node):
        self.latch = node

    def is_start_loop(self):
        return self.startloop

    def is_end_loop(self):
        return self.endloop

    def __repr__(self):
        return str(self)


class Interval(Node):
    def __init__(self, head):
        super(Interval, self).__init__(head.name)
        self.name = 'Interval-%s' % head.name
        self.content = set([head])
        self.end = None
        self.head = head
        head.interval = self

    def __contains__(self, item):
        # If the interval contains only nodes, check if the item is one of them
        if item in self.content:
            return True
        # If the interval contains intervals, we need to check them all
        for nodes in filter(lambda x: isinstance(x, Interval), self.content):
            if item in nodes:
                return True
        return False

    def add_node(self, node):
        if node in self.content:
            return False
        self.content.add(node)
        node.interval = self
        return True

    def compute_end(self, graph):
        for node in self.content:
            for suc in graph.sucs(node):
                if suc not in self.content:
                    self.end = node

    def get_end(self):
        return self.end.get_end()

    def set_next(self, nxt):
        self.head.set_next(nxt.get_head())

    def get_next(self):
        return self.head.get_next()

    def set_loop_type(self, _type):
        self.looptype = _type
        self.get_head().set_loop_type(_type)

    def set_startloop(self):
        self.head.set_startloop()

    def get_head(self):
        return self.head.get_head()

    def __iter__(self):
        for item in self.content:
            yield item

    def __len__(self):
        return len(self.content)

    def __repr__(self):
        return '%s(%s)' % (self.name, self.content)



# This file is part of Androguard.
#
# Copyright (c) 2012 Geoffroy Gueguen <geoffroy.gueguen@gmail.com>
# All Rights Reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

from dad.opcode_ins import INSTRUCTION_SET
from dad.instruction import Variable
from dad.node import Node
from dad.util import log


class BasicBlock(Node):
    def __init__(self, name, block_ins):
        super(BasicBlock, self).__init__(name)
        self.ins = block_ins
        self.ins_range = None
        self.loc_ins = None

    def get_ins(self):
        return self.ins

    def get_loc_with_ins(self):
        if self.loc_ins is None:
            self.loc_ins = zip(range(*self.ins_range), self.ins)
        return self.loc_ins

    def remove_ins(self, loc, ins):
        self.ins.remove(ins)
        self.loc_ins.remove((loc, ins))

    def add_ins(self, new_ins_list):
        for new_ins in new_ins_list:
            self.ins.append(new_ins)

    def number_ins(self, num):
        last_ins_num = num + len(self.ins)
        self.ins_range = (num, last_ins_num)
        self.loc_ins = None
        return last_ins_num

    def process(self, memory, varrs, ind, ifollow, nfollow):
        log('Process not done : %s' % self, 'debug')


class StatementBlock(BasicBlock):
    def __init__(self, name, block_ins):
        super(StatementBlock, self).__init__(name, block_ins)

    def visit(self, visitor):
        return visitor.visit_statement_node(self)

    def __str__(self):
        return '%d-Statement(%s)' % (self.num, self.name)


class ReturnBlock(BasicBlock):
    def __init__(self, name, block_ins):
        super(ReturnBlock, self).__init__(name, block_ins)

    def visit(self, visitor):
        return visitor.visit_return_node(self)

    def __str__(self):
        return '%d-Return(%s)' % (self.num, self.name)


class ThrowBlock(BasicBlock):
    def __init__(self, name, block_ins):
        super(ThrowBlock, self).__init__(name, block_ins)

    def visit(self, visitor):
        return visitor.visit_throw_node(self)

    def __str__(self):
        return '%d-Throw(%s)' % (self.num, self.name)


class SwitchBlock(BasicBlock):
    def __init__(self, name, switch, block_ins):
        super(SwitchBlock, self).__init__(name, block_ins)
        self.switch = switch
        self.cases = []
        self.default = None
        self.node_to_case = {}

    def add_case(self, case):
        self.cases.append(case)

    def visit(self, visitor):
        return visitor.visit_switch_node(self)

    def copy_from(self, node):
        super(SwitchBlock, self).copy_from(node)
        self.cases = node.cases
        self.switch = node.switch

    def update_attribute_with(self, n_map):
        super(SwitchBlock, self).update_attribute_with(n_map)
        self.cases = [n_map.get(n, n) for n in self.cases]
        for n1, n2 in n_map.iteritems():
            if n1 in self.node_to_case:
                self.node_to_case[n2] = self.node_to_case.pop(n1)

    def order_cases(self):
        values = self.switch.get_values()
        if len(values) < len(self.cases):
            self.default = self.cases.pop(0)
        for case, node in zip(values, self.cases):
            self.node_to_case.setdefault(node, []).append(case)

    def __str__(self):
        return '%d-Switch(%s)' % (self.num, self.name)


class CondBlock(BasicBlock):
    def __init__(self, name, block_ins):
        super(CondBlock, self).__init__(name, block_ins)
        self.true = None
        self.false = None

    def set_true(self, node):
        self.true = node

    def set_false(self, node):
        self.false = node

    def update_attribute_with(self, n_map):
        super(CondBlock, self).update_attribute_with(n_map)
        self.true = n_map.get(self.true, self.true)
        self.false = n_map.get(self.false, self.false)

    def neg(self):
        if len(self.ins) > 1:
            raise ('Condition should have only 1 instruction !')
        self.ins[0].neg()

    def visit(self, visitor):
        return visitor.visit_cond_node(self)

    def visit_cond(self, visitor):
        if len(self.ins) > 1:
            raise ('Condition should have only 1 instruction !')
        return visitor.visit_ins(self.ins[0])

    def __str__(self):
        return '%d-If(%s)' % (self.num, self.name)


class Condition(object):
    def __init__(self, cond1, cond2, isand, isnot):
        self.cond1 = cond1
        self.cond2 = cond2
        self.isand = isand
        self.isnot = isnot

    def neg(self):
        self.isand = not self.isand
        self.cond1.neg()
        self.cond2.neg()

    def get_ins(self):
        lins = []
        lins.extend(self.cond1.get_ins())
        lins.extend(self.cond2.get_ins())
        return lins

    def get_loc_with_ins(self):
        loc_ins = []
        loc_ins.extend(self.cond1.get_loc_with_ins())
        loc_ins.extend(self.cond2.get_loc_with_ins())
        return loc_ins

    def visit(self, visitor):
        return visitor.visit_short_circuit_condition(self.isnot, self.isand,
                                             self.cond1, self.cond2)

    def __str__(self):
        if self.isnot:
            ret = '!%s %s %s'
        else:
            ret = '%s %s %s'
        return ret % (self.cond1, ['||', '&&'][self.isand], self.cond2)


class ShortCircuitBlock(CondBlock):
    def __init__(self, name, cond):
        super(ShortCircuitBlock, self).__init__(name, None)
        self.cond = cond

    def get_ins(self):
        return self.cond.get_ins()

    def get_loc_with_ins(self):
        return self.cond.get_loc_with_ins()

    def neg(self):
        self.cond.neg()

    def visit_cond(self, visitor):
        return self.cond.visit(visitor)

    def __str__(self):
        return '%d-SC(%s)' % (self.num, self.cond)


class LoopBlock(CondBlock):
    def __init__(self, name, cond):
        super(LoopBlock, self).__init__(name, None)
        self.cond = cond

    def get_ins(self):
        return self.cond.get_ins()

    def neg(self):
        self.cond.neg()

    def get_loc_with_ins(self):
        return self.cond.get_loc_with_ins()

    def visit(self, visitor):
        return visitor.visit_loop_node(self)

    def visit_cond(self, visitor):
        return self.cond.visit_cond(visitor)

    def update_attribute_with(self, n_map):
        super(LoopBlock, self).update_attribute_with(n_map)
        self.cond.update_attribute_with(n_map)

    def __str__(self):
        if self.looptype.pretest():
            if self.false in self.loop_nodes:
                return '%d-While(!%s)[%s]' % (self.num, self.name, self.cond)
            else:
                return '%d-While(%s)[%s]' % (self.num, self.name, self.cond)
        elif self.looptype.posttest():
            return '%d-DoWhile(%s)[%s]' % (self.num, self.name, self.cond)
        elif self.looptype.endless():
            return '%d-WhileTrue(%s)[%s]' % (self.num, self.name, self.cond)
        return '%dWhileNoType(%s)' % (self.num, self.name)


class TryBlock(BasicBlock):
    def __init__(self, name, block_ins):
        super(TryBlock, self).__init__(name, block_ins)
        self.catch = []

    def add_catch(self, node):
        self.catch.append(node)

    def __str__(self):
        return 'Try(%s)' % self.name


class CatchBlock(BasicBlock):
    def __init__(self, name, block_ins, typeh):
        super(CatchBlock, self).__init__(name, block_ins)
        self.exceptionType = typeh

    def __str__(self):
        return 'Catch(%s)' % self.name


class GenInvokeRetName(object):
    def __init__(self):
        self.num = 0

    def new(self):
        self.num += 1
        self.ret = Variable('tmp%d' % self.num)
        return self.ret

    def set_to(self, ret):
        self.ret = ret

    def last(self):
        return self.ret


def build_node_from_block(block, block_to_node, vmap, gen_ret):
    lins = []
    idx = block.get_start()
    for ins in block.get_instructions():
        opcode = ins.get_op_value()
        if opcode == 0x1f:  # check-cast
            idx += ins.get_length()
            continue
        _ins = INSTRUCTION_SET.get(ins.get_name().lower())
        if _ins is None:
            log('Unknown instruction : %s.' % _ins.get_name().lower(), 'error')
        # fill-array-data
        if opcode == 0x26:
            fillaray = block.get_special_ins(idx)
            lins.append(_ins(ins, vmap, fillaray))
        # invoke-kind[/range]
        elif (0x6e <= opcode <= 0x72 or 0x74 <= opcode <= 0x78):
            lins.append(_ins(ins, vmap, gen_ret))
        # filled-new-array[/range]
        elif 0x24 <= opcode <= 0x25:
            lins.append(_ins(ins, vmap, gen_ret.new()))
        # move-result*
        elif 0xa <= opcode <= 0xc:
            lins.append(_ins(ins, vmap, gen_ret.last()))
        else:
            lins.append(_ins(ins, vmap))
        idx += ins.get_length()
    name = block.get_name()
    # return*
    if 0xe <= opcode <= 0x11:
        node = ReturnBlock(name, lins)
        node.set_return()
    # {packed,sparse}-switch
    elif 0x2b <= opcode <= 0x2c:
        idx -= ins.get_length()
        values = block.get_special_ins(idx)
        node = SwitchBlock(name, values, lins)
        node.set_switch()
    # if-test[z]
    elif 0x32 <= opcode <= 0x3d:
        node = CondBlock(name, lins)
        node.set_cond()
    # throw
    elif opcode == 0x27:
        node = ThrowBlock(name, lins)
        node.set_throw()
    else:
        # goto*
        if 0x28 <= opcode <= 0x2a:
            lins.pop()
        node = StatementBlock(name, lins)
        node.set_stmt()
    block_to_node[block] = node
    return node


def build_exception(node, block, block_to_node):
    log('Exceptions :', 'debug')
    for exception in block.exception_analysis.exceptions:
        log('  => %s' % exception, 'debug')
        catch_node = block_to_node.get(exception[-1])
        if catch_node is None:
            catch_node = Node(block.get_name())
            catch_block = CatchBlock(catch_node, exception[-1],
                                    exception[0])
            block_to_node[exception[-1]] = catch_node
            catch_node.set_content(catch_block)
            #nodegraph.append(catchNode)
        catch_node.num += 1
        node.add_catch(catch_node)
        node.add_node(catch_node)



# This file is part of Androguard.
#
# Copyright (c) 2012 Geoffroy Gueguen <geoffroy.gueguen@gmail.com>
# All Rights Reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

TYPE_DESCRIPTOR = {
    'V': 'void',
    'Z': 'boolean',
    'B': 'byte',
    'S': 'short',
    'C': 'char',
    'I': 'int',
    'J': 'long',
    'F': 'float',
    'D': 'double',
    'STR': 'String',
    'StringBuilder': 'String',
}

ACCESS_FLAGS_CLASSES = {
    0x1:    'public',
    0x2:    'private',
    0x4:    'protected',
    0x8:    'static',
    0x10:   'final',
    0x200:  'interface',
    0x400:  'abstract',
    0x1000: 'synthetic',
    0x2000: 'annotation',
    0x4000: 'enum',
}

ACCESS_FLAGS_FIELDS = {
    0x1:    'public',
    0x2:    'private',
    0x4:    'protected',
    0x8:    'static',
    0x10:   'final',
    0x40:   'volatile',
    0x80:   'transient',
    0x1000: 'synthetic',
    0x4000: 'enum',
}

ACCESS_FLAGS_METHODS = {
    0x1:     'public',
    0x2:     'private',
    0x4:     'protected',
    0x8:     'static',
    0x10:    'final',
    0x20:    'synchronized',
    0x40:    'bridge',
    0x80:    'varargs',
    0x100:   'native',
    0x400:   'abstract',
    0x800:   'strict',
    0x1000:  'synthetic',
    # ACC_CONSTRUCTOR
    0x10000: '',
    0x20000: 'synchronized',
}

TYPE_LEN = {
    'J': 2,
    'D': 2,
}

DEBUG_MODES = {
    'off': (-1),
    'error': 0,
    'log':   1,
    'debug': 2,
}

DEBUG_LEVEL = 'log'
INCREASE_STACK_RECURSION_LIMIT = False


class PprintStream(object):
    '''
    Create an object to be used by pprint
    '''
    def __init__(self):
        self.val = []

    def write(self, obj):
        self.val.append(obj)

    def clean(self):
        self.val = []

    def __str__(self):
        return ''.join(self.val)


def build_path(graph, node1, node2, path=None):
    '''
    Build the path from node1 to node2.
    The path is composed of all the nodes between node1 and node2,
    node1 excluded. Although if there is a loop starting from node1, it will be
    included in the path.
    '''
    if path is None:
        path = []
    if node1 is node2:
        return path
    path.append(node2)
    for pred in graph.preds(node2):
        if pred in path:
            continue
        build_path(graph, node1, pred, path)
    return path


def common_dom(idom, cur, pred):
    if not (cur and pred):
        return cur or pred
    while cur is not pred:
        while cur.num < pred.num:
            pred = idom[pred]
        while cur.num > pred.num:
            cur = idom[cur]
    return cur


def merge_inner(clsdict):
    '''
    Merge the inner class(es) of a class :
    e.g class A { ... } class A$foo{ ... } class A$bar{ ... }
       ==> class A { class foo{...} class bar{...} ... }
    '''
    samelist = False
    done = {}
    while not samelist:
        samelist = True
        classlist = clsdict.keys()
        for classname in classlist:
            parts_name = classname.rsplit('$', 1)
            if len(parts_name) > 1:
                mainclass, innerclass = parts_name
                innerclass = innerclass[:-1]  # remove ';' of the name
                mainclass += ';'
                if mainclass in clsdict:
                    clsdict[mainclass].add_subclass(innerclass,
                                                    clsdict[classname])
                    clsdict[classname].name = innerclass
                    done[classname] = clsdict[classname]
                    del clsdict[classname]
                    samelist = False
                elif mainclass in done:
                    cls = done[mainclass]
                    cls.add_subclass(innerclass, clsdict[classname])
                    clsdict[classname].name = innerclass
                    done[classname] = done[mainclass]
                    del clsdict[classname]
                    samelist = False


def get_type_size(param):
    '''
    Return the number of register needed by the type @param
    '''
    return TYPE_LEN.get(param, 1)


def get_type(atype, size=None):
    '''
    Retrieve the java type of a descriptor (e.g : I)
    '''
    res = TYPE_DESCRIPTOR.get(atype)
    if res is None:
        if atype[0] == 'L':
            if atype.startswith('Ljava/lang'):
                res = atype[1:-1].lstrip('java/lang/').replace('/', '.')
            else:
                res = atype[1:-1].replace('/', '.')
        elif atype[0] == '[':
            if size is None:
                res = '%s[]' % get_type(atype[1:])
            else:
                res = '%s[%s]' % (get_type(atype[1:]), size)
        else:
            res = atype
            log('Unknown descriptor: "%s".' % atype, 'debug')
    return res


def get_params_type(descriptor):
    '''
    Return the parameters type of a descriptor (e.g (IC)V)
    '''
    params = descriptor.split(')')[0][1:].split()
    if params:
        return [param for param in params]
    return []


def create_png(basicblocks, graph, dir_name='graphs2'):
    meth = basicblocks[0].get_method()
    m_name = ''.join([x for x in meth.get_name() if x.isalnum()])
    name = meth.get_class_name().split('/')[-1][:-1] + '#' + m_name
    graph.draw(name, dir_name)


def log(string, mode):
    def _log(log_string):
        print '%s' % log_string

    def _log_debug(dbg_str):
        print 'DEBUG: %s' % dbg_str

    def _log_error(err_str):
        exit('ERROR: %s' % err_str)
    if mode is None:
        return
    mode = DEBUG_MODES.get(mode)
    if mode <= DEBUG_MODES[DEBUG_LEVEL]:
        if mode == DEBUG_MODES['log']:
            _log(string)
        elif mode == DEBUG_MODES['debug']:
            _log_debug(string)
        elif mode == DEBUG_MODES['error']:
            _log_error(string)



# This file is part of Androguard.
#
# Copyright (c) 2012 Geoffroy Gueguen <geoffroy.gueguen@gmail.com>
# All Rights Reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

import sys
sys.path.append('./')

from androguard.core.androgen import AndroguardS
from androguard.core.analysis import analysis
from dad.graph import construct
from dad import util
from dad.dataflow import (immediate_dominator, build_def_use,
                          dead_code_elimination, register_propagation)
from dad.control_flow import identify_structures
from dad.instruction import Param, ThisParam
from dad.writer import Writer


class DvMethod():
    def __init__(self, methanalysis):
        self.method = methanalysis.get_method()
        self.metha = methanalysis
        self.name = self.method.get_name()
        self.lparams = []
        self.basic_blocks = [bb for bb in methanalysis.basic_blocks.get()]
        self.var_to_name = {}
        self.writer = None

        access = self.method.get_access_flags()
        self.access = [flag for flag in util.ACCESS_FLAGS_METHODS
                                     if flag & access]
        desc = self.method.get_descriptor()
        self.type = util.get_type(desc.split(')')[-1])
        self.params_type = util.get_params_type(desc)

        self.exceptions = methanalysis.exceptions.exceptions

        code = self.method.get_code()
        if code is None:
            util.log('No code : %s %s' % (self.name,
                                            self.method.get_class_name()),
                                            'debug')
        else:
            start = code.registers_size - code.ins_size
            if 0x8 not in self.access:
                self.var_to_name[start] = ThisParam(start, self.name)
                self.lparams.append(start)
                start += 1
            num_param = 0
            for ptype in self.params_type:
                param = start + num_param
                self.lparams.append(param)
                self.var_to_name.setdefault(param, Param(param, ptype))
                num_param += util.get_type_size(ptype)

    def process(self):
        util.log('METHOD : %s' % self.name, 'debug')
        if 0:
            from androguard.core import bytecode
            bytecode.method2png('/tmp/graphs/%s#%s.png' % \
                (self.method.get_class_name().split('/')[-1][:-1], self.name),
                                                        self.metha)

        graph = construct(self.basic_blocks, self.var_to_name, self.exceptions)
        self.graph = graph
        if graph is None:
            return

        if 0:
            util.create_png(self.basic_blocks, graph, '/tmp/blocks')
                                                #'dad_graphs/blocks')

        defs, uses = build_def_use(graph, self.lparams)
        dead_code_elimination(graph, uses, defs)
        register_propagation(graph, uses, defs)

        # After the DCE pass, some nodes may be empty, so we can simplify the
        # graph to delete these nodes.
        # We start by restructuring the the graph by spliting the conditional
        # nodes into a pre-header and a header part.
        # We then simplify the graph by merging multiple statement nodes into
        # a single statement node when possible. This also delete empty nodes.
        graph.split_if_nodes()
        graph.simplify()
        graph.reset_rpo()

        idoms = immediate_dominator(graph)
        identify_structures(graph, idoms)

        if 0:
            util.create_png(self.basic_blocks, graph, '/tmp/structured')
                                               #     'dad_graphs/structured')

        self.writer = Writer(graph, self)
        self.writer.write_method()

    def show_source(self):
        if self.writer is not None:
            print self.writer

    def get_source(self):
        if self.writer is not None:
            return '%s' % self.writer
        return ''

    def __repr__(self):
        return 'Method %s' % self.name


class DvClass():
    def __init__(self, dvclass, vma):
        name = dvclass.get_name()
        if name.find('/') > 0:
            pckg, name = name.rsplit('/', 1)
        else:
            pckg, name = '', name
        self.package = pckg[1:].replace('/', '.')
        self.name = name[:-1]

        self.methods = {meth.get_method_idx(): DvMethod(vma.get_method(meth))
                                            for meth in dvclass.get_methods()}
        self.fields = {}
        for field in dvclass.get_fields():
            self.fields[field.get_name()] = field
        self.subclasses = {}
        self.code = []
        self.inner = False

        access = dvclass.get_access_flags()
        self.access = [util.ACCESS_FLAGS_CLASSES.get(flag) for flag in
                            util.ACCESS_FLAGS_CLASSES if flag & access]
        self.prototype = '%s class %s' % (' '.join(self.access), self.name)

        self.interfaces = dvclass.interfaces
        self.superclass = dvclass.get_superclassname()

        util.log('Class : %s' % self.name, 'log')
        util.log('Methods added :', 'log')
        for index, meth in self.methods.iteritems():
            util.log('%s (%s, %s)' % (index, meth.method.get_class_name(),
                                   meth.name), 'log')
        util.log('', 'log')

    def add_subclass(self, innername, dvclass):
        self.subclasses[innername] = dvclass
        dvclass.inner = True

    def get_methods(self, meths=None):
        if meths is None:
            meths = self.methods.copy()
        for klass in self.subclasses.values():
            meths.update(klass.get_methods())
        return meths

    def process_method(self, num):
        meths = self.get_methods()
        if num in meths:
            meths[num].process()
        else:
            util.log('Method %s not found.' % num, 'error')

    def get_source(self):
        source = []
        if not self.inner and self.package:
            source.append('package %s;\n' % self.package)

        if self.superclass is not None:
            self.superclass = self.superclass[1:-1].replace('/', '.')
            if self.superclass.split('.')[-1] == 'Object':
                self.superclass = None
            if self.superclass is not None:
                self.prototype += ' extends %s' % self.superclass
        if self.interfaces is not None:
            self.interfaces = self.interfaces[1:-1].split(' ')
            self.prototype += ' implements %s' % ', '.join(
                        [n[1:-1].replace('/', '.') for n in self.interfaces])

        source.append('%s {\n' % self.prototype)
        for field in self.fields.values():
            access = [util.ACCESS_FLAGS_FIELDS.get(flag) for flag in
                util.ACCESS_FLAGS_FIELDS if flag & field.get_access_flags()]
            f_type = util.get_type(field.get_descriptor())
            name = field.get_name()
            source.append('    %s %s %s;\n' % (' '.join(access), f_type, name))

        for klass in self.subclasses.values():
            source.append(klass.get_source())

        for num, method in self.methods.iteritems():
            source.append(method.get_source())
        source.append('}\n')
        return ''.join(source)

    def show_source(self):
        if not self.inner and self.package:
            print 'package %s;\n' % self.package

        if self.superclass is not None:
            self.superclass = self.superclass[1:-1].replace('/', '.')
            if self.superclass.split('.')[-1] == 'Object':
                self.superclass = None
            if self.superclass is not None:
                self.prototype += ' extends %s' % self.superclass
        if self.interfaces is not None:
            self.interfaces = self.interfaces[1:-1].split(' ')
            self.prototype += ' implements %s' % ', '.join(
                        [n[1:-1].replace('/', '.') for n in self.interfaces])

        print '%s {\n' % self.prototype
        for field in self.fields.values():
            access = [util.ACCESS_FLAGS_FIELDS.get(flag) for flag in
                util.ACCESS_FLAGS_FIELDS if flag & field.get_access_flags()]
            f_type = util.get_type(field.get_descriptor())
            name = field.get_name()
            print '    %s %s %s;\n' % (' '.join(access), f_type, name)

        for klass in self.subclasses.values():
            klass.show_source()

        for num, method in self.methods.iteritems():
            method.show_source()
        print '}\n'

    def process(self):
        for klass in self.subclasses.values():
            klass.process()
        for meth in self.methods:
            self.process_method(meth)

    def __str__(self):
        return 'Class name : %s.' % self.name

    def __repr__(self):
        if not self.subclasses:
            return 'Class(%s)' % self.name
        return 'Class(%s) -- Subclasses(%s)' % (self.name, self.subclasses)


class DvMachine():
    def __init__(self, name):
        vm = AndroguardS(name).get_vm()
        vma = analysis.uVMAnalysis(vm)
        self.classes = {dvclass.get_name(): DvClass(dvclass, vma)
                                        for dvclass in vm.get_classes()}
        util.merge_inner(self.classes)

    def process(self):
        for name, klass in self.classes.iteritems():
            util.log('Processing class: %s' % name, 'log')
            self.classes[name].process()

    def get_classes(self):
        return self.classes.keys()

    def get_class(self, class_name):
        for name, klass in self.classes.iteritems():
            if class_name in name:
                return self.classes[name]

    def show_source(self):
        for klass in self.classes.values():
            klass.show_source()


if __name__ == '__main__':
    # Uncomment to increase the size of the stack.
    # I don't know why, but if the block is `activated` in an if-statement it
    # doesn't work.
    """
    from resource import setrlimit, RLIMIT_STACK
    setrlimit(RLIMIT_STACK, (2 ** 29, -1))
    sys.setrecursionlimit(10 ** 6)
    """

    FILE = 'examples/android/TestsAndroguard/bin/classes.dex'
    if len(sys.argv) > 1:
        MACHINE = DvMachine(sys.argv[1])
    else:
        MACHINE = DvMachine(FILE)

    from pprint import pprint
    TMP = util.PprintStream()
    util.log('===========================', 'log')
    util.log('Classes :', 'log')
    pprint(MACHINE.get_classes(), TMP)
    util.log(TMP, 'log')
    util.log('===========================', 'log')

    CLS_NAME = raw_input('Choose a class: ')
    if CLS_NAME == '*':
        MACHINE.process()
        util.log('\n\nSource:', 'log')
        util.log('===========================', 'log')
        MACHINE.show_source()
        util.log('===========================', 'log')
    else:
        CLS = MACHINE.get_class(CLS_NAME)
        if CLS is None:
            util.log('%s not found.' % CLS_NAME, 'error')
        else:
            util.log('======================', 'log')
            TMP.clean()
            pprint(CLS.get_methods(), TMP)
            util.log(TMP, 'log')
            util.log('======================', 'log')
            METH = raw_input('Method: ')
            if METH == '*':
                util.log('CLASS = %s' % CLS, 'log')
                CLS.process()
            else:
                CLS.process_method(int(METH))
            util.log('\n\nSource:', 'log')
            util.log('===========================', 'log')
            CLS.show_source()



# This file is part of Androguard.
#
# Copyright (c) 2012 Geoffroy Gueguen <geoffroy.gueguen@gmail.com>
# All Rights Reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

from dad.basic_blocks import Condition, ShortCircuitBlock, LoopBlock
from dad.node import Interval
from dad.util import log, common_dom
from dad.graph import Graph


def intervals(graph):
    '''
    Compute the intervals of the graph
    Returns
        interval_graph: a graph of the intervals of G
        interv_of_head: a dict of (header node, interval)
    '''
    interval_graph = Graph()  # graph of intervals
    heads = set([graph.get_entry()])  # set of header nodes
    interv_of_head = {}  # interv_of_head[i] = interval of header i
    processed = dict([(i, False) for i in graph])
    edges = {}

    while heads:
        n = heads.pop()

        if not processed[n]:
            processed[n] = True
            interv_of_head[n] = Interval(n)

            # Check if if there is a node which has all its predecessor in the
            # current interval. If there is, add that node to the interval and
            # repeat until all the possible nodes have been added.
            change = True
            while change:
                change = False
                for m in graph.get_rpo()[1:]:
                    if all([p in interv_of_head[n] for p in graph.preds(m)]):
                        change |= interv_of_head[n].add_node(m)

            # At this stage, a node which is not in the interval, but has one
            # of its predecessor in it, is the header of another interval. So
            # we add all such nodes to the header list.
            for m in graph:
                if m not in interv_of_head[n] and m not in heads:
                    if any([p in interv_of_head[n] for p in graph.preds(m)]):
                        edges.setdefault(interv_of_head[n], []).append(m)
                        heads.add(m)

            interval_graph.add_node(interv_of_head[n])
            interv_of_head[n].compute_end(graph)

    # Edges is a mapping of 'Interval -> [header nodes of interval successors]'
    for e1, heads in edges.items():
        for e2 in heads:
            interval_graph.add_edge(e1, interv_of_head[e2])

    interval_graph.set_entry(graph.get_entry().interval)
    graph_exit = graph.get_exit()
    if graph_exit:
        interval_graph.set_exit(graph_exit.interval)

    return interval_graph, interv_of_head


def derived_sequence(graph):
    '''
    Compute the derived sequence of the graph G
    The intervals of G are collapsed into nodes, intervals of these nodes are
    built, and the process is repeated iteratively until we obtain a single
    node (if the graph is not irreducible)
    '''
    deriv_seq = [graph]
    deriv_interv = []
    single_node = False

    while not single_node:

        interv_graph, interv_of_head = intervals(graph)
        deriv_interv.append(interv_of_head)

        single_node = len(interv_graph) == 1
        if not single_node:
            deriv_seq.append(interv_graph)

        graph = interv_graph
        if 0:
            graph.draw(graph.entry.name, 'dad_graphs/intervals')
        graph.compute_rpo()

    return deriv_seq, deriv_interv


def mark_loop(graph, start, end, L):
    log('MARKLOOP : %s END : %s' % (start, end), 'debug')

    def mark_loop_rec(end):
        if not end in nodes_in_loop:
            nodes_in_loop.append(end)
            for node in graph.preds(end):
                if headnode.num < node.num <= endnode.num:
                    if node in L[start]:
                        mark_loop_rec(node)

    headnode = start.get_head()
    endnode = end.get_end()
    nodes_in_loop = [headnode]
    mark_loop_rec(endnode)
    headnode.set_start_loop()
    endnode.set_end_loop()
    headnode.set_latch_node(endnode)
    return nodes_in_loop


def loop_type(graph, start, end, nodes_in_loop):
    start = start.get_head()
    end = end.get_end()
    if end.is_cond():
        if start.is_cond():
            if start.true in nodes_in_loop and start.false in nodes_in_loop:
                start.set_loop_posttest()
            else:
                start.set_loop_pretest()
        else:
            start.set_loop_posttest()
    else:
        if start.is_cond():
            if start.true in nodes_in_loop and start.false in nodes_in_loop:
                start.set_loop_endless()
            else:
                start.set_loop_pretest()
        else:
            start.set_loop_endless()


def loop_follow(graph, start, end, nodes_in_loop):
    head = start.get_head()
    follow = None
    if start.looptype.pretest():
        if head.true in nodes_in_loop:
            follow = head.false
        else:
            follow = head.true
    elif start.looptype.posttest():
        if end.true in nodes_in_loop:
            follow = end.false
        else:
            follow = end.true
    else:
        num_next = NotImplemented  # Hack to do: num_next = infinity
        for node in nodes_in_loop:
            if node.is_cond():
                if (node.true.num < num_next
                        and node.true not in nodes_in_loop):
                    follow = node.true
                    num_next = follow.num
                elif (node.false.num < num_next
                        and node.false not in nodes_in_loop):
                    follow = node.false
                    num_next = follow.num
    head.set_loop_follow(follow)
    for node in nodes_in_loop:
        node.set_loop_follow(follow)
    log('Start of loop %s' % head, 'debug')
    log('Follow of loop: %s' % head.get_loop_follow(), 'debug')


def loop_struct(Gi, Li):
    first_graph = Gi[0]
    for i, graph in enumerate(Gi):
        L = Li[i]
        for head in sorted(L.keys(), key=lambda x: x.num):
            loop_nodes = set()
            for node in graph.preds(head):
                if node.interval is head.interval:
                    loop_nodes.update(mark_loop(first_graph, head, node, L))
                    head.get_head().set_loop_nodes(loop_nodes)
#                    loop_type(first_graph, head, node, loop_nodes)
#                    loop_follow(first_graph, head, node, loop_nodes)


def if_struct(graph, idoms):
    unresolved = set()
    for node in graph.get_rpo()[::-1]:
        if node.is_cond():
            ldominates = []
            for n, idom in idoms.iteritems():
                if node is idom and len(graph.preds(n)) > 1:
                    ldominates.append(n)
            if len(ldominates) > 0:
                n = max(ldominates, key=lambda x: x.num)
                node.set_if_follow(n)
                for x in unresolved.copy():
                    if node.num < x.num < n.num:
                        x.set_if_follow(n)
                        unresolved.remove(x)
            else:
                unresolved.add(node)
    return unresolved


def switch_struct(graph, idoms):
    unresolved = set()
    for node in graph.post_order():
        if node.is_switch():
            m = node
            for suc in graph.sucs(node):
                if idoms[suc] is not node:
                    m = common_dom(idoms, node, suc)
            ldominates = []
            for n, dom in idoms.iteritems():
                if m is dom and len(graph.preds(n)) > 1:
                    ldominates.append(n)
            if len(ldominates) > 0:
                n = max(ldominates, key=lambda x: x.num)
                node.set_switch_follow(n)
                for x in unresolved:
                    x.set_switch_follow(n)
                unresolved = set()
            else:
                unresolved.add(node)
            node.order_cases()


def short_circuit_struct(graph, idom, node_map):
    def MergeNodes(node1, node2, is_and, is_not):
        lpreds = set()
        ldests = set()
        for node in (node1, node2):
            lpreds.update(graph.preds(node))
            ldests.update(graph.sucs(node))
            graph.remove_node(node)
            done.add(node)
        lpreds.difference_update((node1, node2))
        ldests.difference_update((node1, node2))

        entry = graph.get_entry() in (node1, node2)

        new_name = '%s+%s' % (node1.name, node2.name)
        condition = Condition(node1, node2, is_and, is_not)

        new_node = ShortCircuitBlock(new_name, condition)
        for old_n, new_n in node_map.iteritems():
            if new_n in (node1, node2):
                node_map[old_n] = new_node
        node_map[node1] = new_node
        node_map[node2] = new_node
        idom[new_node] = idom[node1]
        idom.pop(node1)
        idom.pop(node2)
        new_node.copy_from(node1)

        graph.add_node(new_node)

        for pred in lpreds:
            pred.update_attribute_with(node_map)
            graph.add_edge(node_map.get(pred, pred), new_node)
        for dest in ldests:
            graph.add_edge(new_node, node_map.get(dest, dest))
        if entry:
            graph.set_entry(new_node)
        return new_node

    change = True
    while change:
        change = False
        done = set()
        for node in graph.post_order():
            if node.is_cond() and node not in done:
                then = node.true
                els = node.false
                if node in (then, els):
                    continue
                if then.is_cond() and len(graph.preds(then)) == 1:
                    if then.false is els:  # node && t
                        change = True
                        merged_node = MergeNodes(node, then, True, False)
                        merged_node.set_true(then.true)
                        merged_node.set_false(els)
                    elif then.true is els:  # !node || t
                        change = True
                        merged_node = MergeNodes(node, then, False, True)
                        merged_node.set_true(els)
                        merged_node.set_false(then.false)
                elif els.is_cond() and len(graph.preds(els)) == 1:
                    if els.false is then:  # !node && e
                        change = True
                        merged_node = MergeNodes(node, els, True, True)
                        merged_node.set_true(els.true)
                        merged_node.set_false(then)
                    elif els.true is then:  # node || e
                        change = True
                        merged_node = MergeNodes(node, els, False, False)
                        merged_node.set_true(then)
                        merged_node.set_false(els.false)
            done.add(node)
        if change:
            graph.reset_rpo()


def while_block_struct(graph, node_map):
    change = False
    for node in graph.get_rpo()[:]:
        if node.is_start_loop():
            change = True
            new_node = LoopBlock(node.name, node)
            node_map[node] = new_node
            new_node.copy_from(node)

            entry = node is graph.get_entry()
            lpreds = graph.preds(node)
            lsuccs = graph.sucs(node)

            for pred in lpreds:
                graph.add_edge(node_map.get(pred, pred), new_node)

            for suc in lsuccs:
                    graph.add_edge(new_node, node_map.get(suc, suc))
            if entry:
                graph.set_entry(new_node)

            if node.is_cond():
                new_node.set_true(node.true)
                new_node.set_false(node.false)

            graph.add_node(new_node)
            graph.remove_node(node)

    if change:
        graph.reset_rpo()


def update_dom(idoms, node_map):
    for n, dom in idoms.iteritems():
        idoms[n] = node_map.get(dom, dom)


def identify_structures(graph, idoms):
    Gi, Li = derived_sequence(graph)
    switch_struct(graph, idoms)
    loop_struct(Gi, Li)
    node_map = {}

    short_circuit_struct(graph, idoms, node_map)
    update_dom(idoms, node_map)

    if_unresolved = if_struct(graph, idoms)

    while_block_struct(graph, node_map)
    update_dom(idoms, node_map)

    loop_starts = []
    for node in graph.get_rpo():
        node.update_attribute_with(node_map)
        if node.is_start_loop():
            loop_starts.append(node)
    for node in loop_starts:
        loop_type(graph, node, node.latch, node.loop_nodes)
        loop_follow(graph, node, node.latch, node.loop_nodes)

    for node in if_unresolved:
        follows = [n for n in (node.loop_follow, node.switch_follow) if n]
        if len(follows) >= 1:
            follow = min(follows, key=lambda x: x.num)
            node.set_if_follow(follow)
    del node_map



#!/usr/bin/env python

# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

import sys

from androguard.core import androconf

sys.path.append("./elsim/")
from elsim.elsign import dalvik_elsign

from optparse import OptionParser

option_0 = { 'name' : ('-i', '--input'), 'help' : 'file : use this filename', 'nargs' : 1 }
option_1 = { 'name' : ('-r', '--remove'), 'help' : 'remote the signature', 'nargs' : 1 }
option_2 = { 'name' : ('-o', '--output'), 'help' : 'output database', 'nargs' : 1 }
option_3 = { 'name' : ('-l', '--list'), 'help' : 'list signatures in database', 'nargs' : 1 }
option_4 = { 'name' : ('-c', '--check'), 'help' : 'check signatures in database', 'nargs' : 1 }
option_5 = { 'name' : ('-v', '--version'), 'help' : 'version of the API', 'action' : 'count' }

options = [option_0, option_1, option_2, option_3, option_4, option_5]

def main(options, arguments) :
    s = dalvik_elsign.CSignature(pcs=dalvik_elsign.PublicCSignature)
    if options.input != None :
        ret = s.add_file( open( options.input, "rb" ).read() )
        if ret != None and options.output != None :
            s.add_indb( ret, options.output )

    elif options.list != None :
        s.list_indb( options.list )

    elif options.remove != None :
        s.remove_indb( options.remove, options.output )

    elif options.check != None :
        s.check_db( options.check )

    elif options.version != None :
        print "Androcsign version %s" % androconf.ANDROGUARD_VERSION

if __name__ == "__main__" :
    parser = OptionParser()
    for option in options :
        param = option['name']
        del option['name']
        parser.add_option(*param, **option)

    options, arguments = parser.parse_args()
    sys.argv[:] = arguments
    main(options, arguments)
